"""Honkai battle chronicle component."""

import asyncio
import typing

from genshin import errors, types, utility
from genshin.models.honkai import chronicle as models

from . import base

__all__ = ["HonkaiBattleChronicleClient"]


class HonkaiBattleChronicleClient(base.BaseBattleChronicleClient):
    """Honkai battle chronicle component."""

    async def _request_honkai_record(
        self,
        endpoint: str,
        uid: typing.Optional[int] = None,
        *,
        lang: typing.Optional[str] = None,
        cache: bool = True,
    ) -> typing.Mapping[str, typing.Any]:
        """Get an arbitrary honkai object."""
        uid = uid or await self._get_uid(types.Game.HONKAI)

        cache_key: typing.Optional[base.ChronicleCacheKey] = None
        if cache:
            cache_key = base.ChronicleCacheKey(
                types.Game.HONKAI,
                endpoint,
                uid,
                lang=lang or self.lang,
            )

        return await self.request_game_record(
            endpoint,
            lang=lang,
            game=types.Game.HONKAI,
            region=types.Region.OVERSEAS,
            params=dict(role_id=uid, server=utility.recognize_honkai_server(uid)),
            cache=cache_key,
        )

    async def get_honkai_user(
        self,
        uid: int,
        *,
        lang: typing.Optional[str] = None,
    ) -> models.HonkaiUserStats:
        """Get honkai user stats."""
        data = await self._request_honkai_record("index", uid, lang=lang)
        return models.HonkaiUserStats(**data, lang=lang or self.lang)

    async def get_honkai_battlesuits(
        self,
        uid: int,
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.FullBattlesuit]:
        """Get honkai battlesuits."""
        data = await self._request_honkai_record("characters", uid, lang=lang)
        return [models.FullBattlesuit(**char["character"]) for char in data["characters"]]

    async def get_honkai_old_abyss(
        self,
        uid: int,
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.OldAbyss]:
        """Get honkai old abyss.

        Only for level > 80.
        """
        data = await self._request_honkai_record("latestOldAbyssReport", uid, lang=lang)
        return [models.OldAbyss(**x, lang=lang or self.lang) for x in data["reports"]]

    async def get_honkai_superstring_abyss(
        self,
        uid: int,
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.SuperstringAbyss]:
        """Get honkai superstring abyss.

        Only for level <= 80.
        """
        data = await self._request_honkai_record("newAbyssReport", uid, lang=lang)
        return [models.SuperstringAbyss(**x, lang=lang or self.lang) for x in data["reports"]]

    async def get_honkai_abyss(
        self,
        uid: int,
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[typing.Union[models.SuperstringAbyss, models.OldAbyss]]:
        """Get honkai abyss."""
        possible = await asyncio.gather(
            self.get_honkai_old_abyss(uid, lang=lang),
            self.get_honkai_superstring_abyss(uid, lang=lang),
            return_exceptions=True,
        )
        for abyss in possible:
            if not isinstance(abyss, BaseException):
                return abyss
            if not isinstance(abyss, errors.InternalDatabaseError):
                raise abyss from None

        return []

    async def get_honkai_elysian_realm(
        self,
        uid: int,
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.ElysianRealm]:
        """Get honkai elysian realm."""
        data = await self._request_honkai_record("godWar", uid, lang=lang)
        return [models.ElysianRealm(**x) for x in data["records"]]

    async def get_honkai_memorial_arena(
        self,
        uid: int,
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.MemorialArena]:
        """Get honkai memorial arena."""
        data = await self._request_honkai_record("battleFieldReport", uid, lang=lang)
        return [models.MemorialArena(**x, lang=lang or self.lang) for x in data["reports"]]

    async def get_full_honkai_user(
        self,
        uid: int,
        *,
        lang: typing.Optional[str] = None,
    ) -> models.FullHonkaiUserStats:
        """Get a full honkai user."""
        user, battlesuits, abyss, mr, er = await asyncio.gather(
            self.get_honkai_user(uid, lang=lang),
            self.get_honkai_battlesuits(uid, lang=lang),
            self.get_honkai_abyss(uid, lang=lang),
            self.get_honkai_memorial_arena(uid, lang=lang),
            self.get_honkai_elysian_realm(uid, lang=lang),
        )

        return models.FullHonkaiUserStats(
            **user.dict(),
            battlesuits=battlesuits,
            abyss=abyss,
            memorial_arena=mr,
            elysian_realm=er,
        )

    get_old_abyss = get_honkai_old_abyss
    get_superstring_abyss = get_honkai_superstring_abyss
    get_elysian_realm = get_honkai_elysian_realm
    get_memorial_arena = get_honkai_memorial_arena



"""Battle chronicle component."""

from . import genshin, honkai, starrail

__all__ = ["BattleChronicleClient"]


class BattleChronicleClient(
    genshin.GenshinBattleChronicleClient,
    honkai.HonkaiBattleChronicleClient,
    starrail.StarRailBattleChronicleClient,
):
    """Battle chronicle component."""



"""Calculator builder object.

Over-engineered for the sake of extendability and maintainability.
"""

from __future__ import annotations

import abc
import asyncio
import typing

import genshin.models.genshin as genshin_models
from genshin import types
from genshin.models.genshin import calculator as models

if typing.TYPE_CHECKING:
    from .client import CalculatorClient as Client

__all__ = ["Calculator", "FurnishingCalculator"]

T = typing.TypeVar("T")
CallableT = typing.TypeVar("CallableT", bound="typing.Callable[..., typing.Awaitable[object]]")


def _cache(func: CallableT) -> CallableT:
    """Cache a method."""

    async def wrapper(self: CalculatorState, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        async with self.lock:
            if value := self.cache.get(func.__name__):
                return value

            value = await func(self, *args, **kwargs)
            self.cache[func.__name__] = value

        return value

    return typing.cast("CallableT", wrapper)


class CalculatorState:
    """Stores character details if multiple objects require them."""

    client: Client
    cache: typing.Dict[str, typing.Any]
    lock: asyncio.Lock

    character_id: typing.Optional[int] = None

    def __init__(self, client: Client) -> None:
        self.client = client
        self.cache = {}
        self.lock = asyncio.Lock()

    @_cache
    async def get_character_details(self) -> models.CalculatorCharacterDetails:
        """Get character details."""
        if self.character_id is None:
            raise TypeError("No specified character.")

        return await self.client.get_character_details(self.character_id)

    @_cache
    async def get_character_talents(self) -> typing.Sequence[models.CalculatorTalent]:
        """Get talent ids."""
        if self.character_id is None:
            raise TypeError("No specified character.")

        return await self.client.get_character_talents(self.character_id)

    @_cache
    async def get_artifact_ids(self, artifact_id: int) -> typing.Sequence[int]:
        """Get artifact ids."""
        others = await self.client.get_complete_artifact_set(artifact_id)
        return [artifact_id] + [other.id for other in others]


class CalculatorResolver(abc.ABC, typing.Generic[T]):
    """Auto-resolving calculator object."""

    @abc.abstractmethod
    async def __call__(self, state: CalculatorState) -> T:
        """Resolve the object into concrete data."""


class CharacterResolver(CalculatorResolver[typing.Mapping[str, typing.Any]]):
    def __init__(
        self,
        character: types.IDOr[genshin_models.BaseCharacter],
        current: typing.Optional[int] = None,
        target: typing.Optional[int] = None,
        *,
        element: typing.Optional[int] = None,
    ) -> None:
        if isinstance(character, genshin_models.BaseCharacter):
            current = current or getattr(character, "level", None)
            character = character.id

        self.id = character
        self.current = current
        self.target = target
        self.element = element

    async def __call__(self, state: CalculatorState) -> typing.Mapping[str, typing.Any]:
        if self.current is None or self.target is None:
            return {}

        data = dict(
            avatar_id=self.id,
            avatar_level_current=self.current,
            avatar_level_target=self.target,
        )
        if self.element:
            data.update(element_attr_id=self.element)

        return data


class WeaponResolver(CalculatorResolver[typing.Mapping[str, typing.Any]]):
    id: int
    current: int
    target: int

    def __init__(self, weapon: int, current: int, target: int) -> None:
        self.id = weapon
        self.current = current
        self.target = target

    async def __call__(self, state: CalculatorState) -> typing.Mapping[str, typing.Any]:
        return dict(
            id=self.id,
            level_current=self.current,
            level_target=self.target,
        )


class CurrentWeaponResolver(WeaponResolver):
    id: int
    current: int
    target: int

    def __init__(self, target: int):
        self.target = target

    async def __call__(self, state: CalculatorState) -> typing.Mapping[str, typing.Any]:
        details = await state.get_character_details()
        self.id = details.weapon.id
        self.current = details.weapon.level
        return await super().__call__(state)


class ArtifactResolver(CalculatorResolver[typing.Sequence[typing.Mapping[str, typing.Any]]]):
    data: typing.List[typing.Mapping[str, typing.Any]]

    def __init__(self) -> None:
        self.data = []

    def add_artifact(self, id: int, current: int, target: int) -> None:
        self.data.append(dict(id=id, level_current=current, level_target=target))

    async def __call__(self, state: CalculatorState) -> typing.Sequence[typing.Mapping[str, typing.Any]]:
        return self.data


class ArtifactSetResolver(ArtifactResolver):
    def __init__(self, any_artifact_id: int, current: int, target: int) -> None:
        self.id = any_artifact_id
        self.current = current
        self.target = target

        super().__init__()

    async def __call__(self, state: CalculatorState) -> typing.Sequence[typing.Mapping[str, typing.Any]]:
        artifact_ids = await state.get_artifact_ids(self.id)

        for artifact_id in artifact_ids:
            self.add_artifact(artifact_id, self.current, self.target)

        return self.data


class CurrentArtifactResolver(ArtifactResolver):
    artifacts: typing.Sequence[typing.Optional[int]]

    def __init__(
        self,
        target: typing.Optional[int] = None,
        *,
        flower: typing.Optional[int] = None,
        feather: typing.Optional[int] = None,
        sands: typing.Optional[int] = None,
        goblet: typing.Optional[int] = None,
        circlet: typing.Optional[int] = None,
    ) -> None:
        if target:
            self.artifacts = (target,) * 5
        else:
            self.artifacts = (flower, feather, sands, goblet, circlet)

    async def __call__(self, state: CalculatorState) -> typing.Sequence[typing.Mapping[str, typing.Any]]:
        details = await state.get_character_details()

        for artifact in details.artifacts:
            if target := self.artifacts[artifact.pos - 1]:
                self.add_artifact(artifact.id, artifact.level, target)

        return self.data


class TalentResolver(CalculatorResolver[typing.Sequence[typing.Mapping[str, typing.Any]]]):
    data: typing.List[typing.Mapping[str, typing.Any]]

    def __init__(self) -> None:
        self.data = []

    def add_talent(self, id: int, current: int, target: int) -> None:
        self.data.append(dict(id=id, level_current=current, level_target=target))

    async def __call__(self, state: CalculatorState) -> typing.Sequence[typing.Mapping[str, typing.Any]]:
        return self.data


class CurrentTalentResolver(TalentResolver):
    talents: typing.Mapping[str, typing.Optional[int]]

    def __init__(
        self,
        target: typing.Optional[int] = None,
        current: typing.Optional[int] = None,
        *,
        attack: typing.Optional[int] = None,
        skill: typing.Optional[int] = None,
        burst: typing.Optional[int] = None,
    ) -> None:
        self.current = current
        if target:
            self.talents = {
                "attack": target,
                "skill": target,
                "burst": target,
            }
        else:
            self.talents = {
                "attack": attack,
                "skill": skill,
                "burst": burst,
            }

        super().__init__()

    async def __call__(self, state: CalculatorState) -> typing.Sequence[typing.Mapping[str, typing.Any]]:
        if self.current:
            talents = await state.get_character_talents()
        else:
            details = await state.get_character_details()
            talents = details.talents
            self.current = 0

        if talents[2].type == "dash":
            ordered = (talents[0], talents[1], talents[3])
        else:
            ordered = (talents[0], talents[1], talents[2])

        for talent, name in zip(ordered, ("attack", "skill", "burst")):
            if target := self.talents[name]:
                self.add_talent(talent.group_id, talent.level or self.current, target)

        return self.data


class Calculator:
    """Builder for the genshin impact enhancement calculator."""

    client: Client
    lang: typing.Optional[str]

    character: typing.Optional[CharacterResolver]
    weapon: typing.Optional[WeaponResolver]
    artifacts: typing.Optional[ArtifactResolver]
    talents: typing.Optional[TalentResolver]

    _state: CalculatorState

    def __init__(self, client: Client, *, lang: typing.Optional[str] = None) -> None:
        self.client = client
        self.lang = lang

        self.character = None
        self.weapon = None
        self.artifacts = None
        self.talents = None

        self._state = CalculatorState(client)

    def set_character(
        self,
        character: types.IDOr[genshin_models.BaseCharacter],
        current: typing.Optional[int] = None,
        target: typing.Optional[int] = None,
        *,
        element: typing.Optional[int] = None,
    ) -> Calculator:
        """Set the character."""
        self.character = CharacterResolver(character, current, target, element=element)
        self._state.character_id = self.character.id
        return self

    def set_weapon(self, id: int, current: int, target: int) -> Calculator:
        """Set the weapon."""
        self.weapon = WeaponResolver(id, current, target)
        return self

    def add_artifact(self, id: int, current: int, target: int) -> Calculator:
        """Add an artifact."""
        if type(self.artifacts) is not ArtifactResolver:
            self.artifacts = ArtifactResolver()

        self.artifacts.add_artifact(id, current, target)
        return self

    def set_artifact_set(self, any_artifact_id: int, current: int, target: int) -> Calculator:
        """Set an artifact set."""
        self.artifacts = ArtifactSetResolver(any_artifact_id, current, target)
        return self

    def add_talent(self, group_id: int, current: int, target: int) -> Calculator:
        """Add a talent."""
        if type(self.talents) is not TalentResolver:
            self.talents = TalentResolver()

        self.talents.add_talent(group_id, current, target)
        return self

    def with_current_weapon(self, target: int) -> Calculator:
        """Set the weapon of the selected character."""
        self.weapon = CurrentWeaponResolver(target)
        return self

    def with_current_artifacts(
        self,
        target: typing.Optional[int] = None,
        *,
        flower: typing.Optional[int] = None,
        feather: typing.Optional[int] = None,
        sands: typing.Optional[int] = None,
        goblet: typing.Optional[int] = None,
        circlet: typing.Optional[int] = None,
    ) -> Calculator:
        """Add all artifacts of the selected character."""
        self.artifacts = CurrentArtifactResolver(
            target,
            flower=flower,
            feather=feather,
            sands=sands,
            goblet=goblet,
            circlet=circlet,
        )
        return self

    def with_current_talents(
        self,
        target: typing.Optional[int] = None,
        current: typing.Optional[int] = None,
        *,
        attack: typing.Optional[int] = None,
        skill: typing.Optional[int] = None,
        burst: typing.Optional[int] = None,
    ) -> Calculator:
        """Add all talents of the currently selected character."""
        self.talents = CurrentTalentResolver(
            target=target,
            current=current,
            attack=attack,
            skill=skill,
            burst=burst,
        )
        return self

    async def build(self) -> typing.Mapping[str, typing.Any]:
        """Build the calculator object."""
        data: typing.Dict[str, typing.Any] = {}

        if self.character:
            data.update(await self.character(self._state))

        if self.weapon:
            data["weapon"] = await self.weapon(self._state)

        if self.artifacts:
            data["reliquary_list"] = await self.artifacts(self._state)

        if self.talents:
            data["skill_list"] = await self.talents(self._state)

        return data

    async def calculate(self) -> models.CalculatorResult:
        """Execute the calculator."""
        return await self.client._execute_calculator(await self.build(), lang=self.lang)

    def __await__(self) -> typing.Generator[typing.Any, None, models.CalculatorResult]:
        return self.calculate().__await__()


class FurnishingCalculator:
    """Builder for the genshin impact furnishing calculator."""

    client: Client
    lang: typing.Optional[str]

    furnishings: typing.Dict[int, int]
    replica_code: typing.Optional[int] = None
    replica_region: typing.Optional[str] = None

    def __init__(self, client: Client, *, lang: typing.Optional[str] = None) -> None:
        self.client = client
        self.lang = lang

        self.furnishings = {}
        self.replica_code = None
        self.replica_region = None

    def add_furnishing(self, id: types.IDOr[models.CalculatorFurnishing], amount: int = 1) -> FurnishingCalculator:
        """Add a furnishing."""
        self.furnishings.setdefault(int(id), 0)
        self.furnishings[int(id)] += amount
        return self

    def with_replica(self, code: int, *, region: typing.Optional[str] = None) -> FurnishingCalculator:
        """Set the replica code."""
        self.replica_code = code
        self.replica_region = region
        return self

    async def build(self) -> typing.Mapping[str, typing.Any]:
        """Build the calculator object."""
        data: typing.Dict[str, typing.Any] = {}

        if self.replica_code:
            furnishings = await self.client.get_teapot_replica_blueprint(self.replica_code, region=self.replica_region)
            self.furnishings.update({furnishing.id: furnishing.amount or 1 for furnishing in furnishings})

        data["list"] = [{"id": id, "cnt": amount} for id, amount in self.furnishings.items()]

        return data

    async def calculate(self) -> models.CalculatorFurnishingResults:
        """Execute the calculator."""
        return await self.client._execute_furnishings_calculator(await self.build(), lang=self.lang)

    def __await__(self) -> typing.Generator[typing.Any, None, models.CalculatorFurnishingResults]:
        return self.calculate().__await__()



"""Main auth client."""

import asyncio
import logging
import typing

import aiohttp

from genshin import errors, types
from genshin.client import routes
from genshin.client.components import base
from genshin.client.manager import managers
from genshin.client.manager.cookie import fetch_cookie_token_with_game_token, fetch_stoken_with_game_token
from genshin.models.auth.cookie import (
    AppLoginResult,
    CNWebLoginResult,
    GameLoginResult,
    MobileLoginResult,
    QRLoginResult,
    WebLoginResult,
)
from genshin.models.auth.geetest import MMT, RiskyCheckMMT, RiskyCheckMMTResult, SessionMMT, SessionMMTResult
from genshin.models.auth.qrcode import QRCodeStatus
from genshin.models.auth.verification import ActionTicket
from genshin.types import Game
from genshin.utility import auth as auth_utility
from genshin.utility import ds as ds_utility

from . import server, subclients

__all__ = ["AuthClient"]

LOGGER_ = logging.getLogger(__name__)


class AuthClient(subclients.AppAuthClient, subclients.WebAuthClient, subclients.GameAuthClient):
    """Auth client component."""

    async def login_with_password(
        self,
        account: str,
        password: str,
        *,
        port: int = 5000,
        encrypted: bool = False,
        geetest_solver: typing.Optional[typing.Callable[[SessionMMT], typing.Awaitable[SessionMMTResult]]] = None,
    ) -> typing.Union[WebLoginResult, CNWebLoginResult]:
        """Login with a password via web endpoint.

        Endpoint is chosen based on client region.

        Note that this will start a webserver if captcha is
        triggered and `geetest_solver` is not passed.

        Raises
        ------
        - AccountLoginFail: Invalid password provided.
        - AccountDoesNotExist: Invalid email/username.
        """
        if self.region is types.Region.CHINESE:
            return await self.cn_login_with_password(
                account, password, encrypted=encrypted, port=port, geetest_solver=geetest_solver
            )

        return await self.os_login_with_password(
            account, password, port=port, encrypted=encrypted, geetest_solver=geetest_solver
        )

    @base.region_specific(types.Region.OVERSEAS)
    async def os_login_with_password(
        self,
        account: str,
        password: str,
        *,
        port: int = 5000,
        encrypted: bool = False,
        token_type: typing.Optional[int] = 6,
        geetest_solver: typing.Optional[typing.Callable[[SessionMMT], typing.Awaitable[SessionMMTResult]]] = None,
    ) -> WebLoginResult:
        """Login with a password via web endpoint.

        Note that this will start a webserver if captcha is
        triggered and `geetest_solver` is not passed.

        Raises
        ------
        - AccountLoginFail: Invalid password provided.
        - AccountDoesNotExist: Invalid email/username.
        """
        result = await self._os_web_login(account, password, encrypted=encrypted, token_type=token_type)

        if not isinstance(result, SessionMMT):
            # Captcha not triggered
            return result

        if geetest_solver:
            mmt_result = await geetest_solver(result)
        else:
            mmt_result = await server.solve_geetest(result, port=port)

        return await self._os_web_login(
            account, password, encrypted=encrypted, token_type=token_type, mmt_result=mmt_result
        )

    @base.region_specific(types.Region.CHINESE)
    async def cn_login_with_password(
        self,
        account: str,
        password: str,
        *,
        encrypted: bool = False,
        port: int = 5000,
        geetest_solver: typing.Optional[typing.Callable[[SessionMMT], typing.Awaitable[SessionMMTResult]]] = None,
    ) -> CNWebLoginResult:
        """Login with a password via Miyoushe loginByPassword endpoint.

        Note that this will start a webserver if captcha is
        triggered and `geetest_solver` is not passed.
        """
        result = await self._cn_web_login(account, password, encrypted=encrypted)

        if not isinstance(result, SessionMMT):
            # Captcha not triggered
            return result

        if geetest_solver:
            mmt_result = await geetest_solver(result)
        else:
            mmt_result = await server.solve_geetest(result, port=port)

        return await self._cn_web_login(account, password, encrypted=encrypted, mmt_result=mmt_result)

    @base.region_specific(types.Region.OVERSEAS)
    async def check_mobile_number_validity(self, mobile: str) -> bool:
        """Check if a mobile number is valid (it's registered on Miyoushe).

        Returns True if the mobile number is valid, False otherwise.
        """
        async with aiohttp.ClientSession() as session:
            async with session.get(
                routes.CHECK_MOBILE_VALIDITY_URL.get_url(),
                params={"mobile": mobile},
            ) as r:
                data = await r.json()

        return data["data"]["status"] != data["data"]["is_registable"]

    @base.region_specific(types.Region.CHINESE)
    async def login_with_mobile_number(
        self,
        mobile: str,
        *,
        encrypted: bool = False,
        port: int = 5000,
    ) -> MobileLoginResult:
        """Login with mobile number, returns cookies.

        Only works for Chinese region (Miyoushe) users, do not include
        area code (+86) in the mobile number.

        Steps:
        1. Sends OTP to the provided mobile number.
        2. If captcha is triggered, prompts the user to solve it.
        3. Lets user enter the OTP.
        4. Logs in with the OTP.
        5. Returns cookies.
        """
        result = await self._send_mobile_otp(mobile, encrypted=encrypted)

        if isinstance(result, SessionMMT):
            # Captcha triggered
            mmt_result = await server.solve_geetest(result, port=port)
            await self._send_mobile_otp(mobile, encrypted=encrypted, mmt_result=mmt_result)

        otp = await server.enter_code(port=port)
        return await self._login_with_mobile_otp(mobile, otp, encrypted=encrypted)

    @base.region_specific(types.Region.OVERSEAS)
    async def login_with_app_password(
        self,
        account: str,
        password: str,
        *,
        encrypted: bool = False,
        port: int = 5000,
        geetest_solver: typing.Optional[typing.Callable[[SessionMMT], typing.Awaitable[SessionMMTResult]]] = None,
    ) -> AppLoginResult:
        """Login with a password via HoYoLab app endpoint.

        Note that this will start a webserver if either of the
        following happens:

        1. Captcha is triggered and `geetest_solver` is not passed.
        2. Email verification is triggered (can happen if you
        first login with a new device).

        Raises
        ------
        - AccountLoginFail: Invalid password provided.
        - AccountDoesNotExist: Invalid email/username.
        - VerificationCodeRateLimited: Too many verification code requests.
        """
        result = await self._app_login(account, password, encrypted=encrypted)

        if isinstance(result, SessionMMT):
            # Captcha triggered
            if geetest_solver:
                mmt_result = await geetest_solver(result)
            else:
                mmt_result = await server.solve_geetest(result, port=port)

            result = await self._app_login(account, password, encrypted=encrypted, mmt_result=mmt_result)

        if isinstance(result, ActionTicket):
            # Email verification required
            mmt = await self._send_verification_email(result)
            if mmt:
                if geetest_solver:
                    mmt_result = await geetest_solver(mmt)
                else:
                    mmt_result = await server.solve_geetest(mmt, port=port)

                await asyncio.sleep(2)  # Add delay to prevent [-3206]
                await self._send_verification_email(result, mmt_result=mmt_result)

            code = await server.enter_code(port=port)
            await self._verify_email(code, result)

            result = await self._app_login(account, password, encrypted=encrypted, ticket=result)

        return result

    @base.region_specific(types.Region.CHINESE)
    async def login_with_qrcode(self) -> QRLoginResult:
        """Login with QR code, only available for Miyoushe users."""
        import qrcode
        import qrcode.image.pil
        from qrcode.constants import ERROR_CORRECT_L

        creation_result = await self._create_qrcode()
        qrcode_: qrcode.image.pil.PilImage = qrcode.make(creation_result.url, error_correction=ERROR_CORRECT_L)  # type: ignore
        qrcode_.show()

        scanned = False
        while True:
            check_result = await self._check_qrcode(
                creation_result.app_id, creation_result.device_id, creation_result.ticket
            )
            if check_result.status == QRCodeStatus.SCANNED and not scanned:
                LOGGER_.info("QR code scanned")
                scanned = True
            elif check_result.status == QRCodeStatus.CONFIRMED:
                LOGGER_.info("QR code login confirmed")
                break

            await asyncio.sleep(2)

        raw_data = check_result.payload.raw
        assert raw_data is not None

        cookie_token = await fetch_cookie_token_with_game_token(
            game_token=raw_data.game_token, account_id=raw_data.account_id
        )
        stoken = await fetch_stoken_with_game_token(game_token=raw_data.game_token, account_id=int(raw_data.account_id))

        cookies = {
            "stoken_v2": stoken.token,
            "ltuid": stoken.aid,
            "account_id": stoken.aid,
            "ltmid": stoken.mid,
            "cookie_token": cookie_token,
        }
        self.set_cookies(cookies)
        return QRLoginResult(**cookies)

    @base.region_specific(types.Region.CHINESE)
    @managers.no_multi
    async def create_mmt(self) -> MMT:
        """Create a geetest challenge."""
        is_genshin = self.game is Game.GENSHIN
        headers = {
            "DS": ds_utility.generate_create_geetest_ds(),
            "x-rpc-challenge_game": "2" if is_genshin else "6",
            "x-rpc-page": "v4.1.5-ys_#ys" if is_genshin else "v1.4.1-rpg_#/rpg",
            "x-rpc-tool-verison": "v4.1.5-ys" if is_genshin else "v1.4.1-rpg",
            **auth_utility.CREATE_MMT_HEADERS,
        }

        assert isinstance(self.cookie_manager, managers.CookieManager)
        async with self.cookie_manager.create_session() as session:
            async with session.get(
                routes.CREATE_MMT_URL.get_url(), headers=headers, cookies=self.cookie_manager.cookies
            ) as r:
                data = await r.json()

        if not data["data"]:
            errors.raise_for_retcode(data)

        return MMT(**data["data"])

    @base.region_specific(types.Region.OVERSEAS)
    async def os_game_login(
        self,
        account: str,
        password: str,
        *,
        encrypted: bool = False,
        port: int = 5000,
        geetest_solver: typing.Optional[typing.Callable[[RiskyCheckMMT], typing.Awaitable[RiskyCheckMMTResult]]] = None,
    ) -> GameLoginResult:
        """Perform a login to the game.

        Raises
        ------
        - IncorrectGameAccount: Invalid account provided.
        - IncorrectGamePassword: Invalid password provided.
        """
        result = await self._shield_login(account, password, encrypted=encrypted)

        if isinstance(result, RiskyCheckMMT):
            if geetest_solver:
                mmt_result = await geetest_solver(result)
            else:
                mmt_result = await server.solve_geetest(result, port=port)

            result = await self._shield_login(account, password, encrypted=encrypted, mmt_result=mmt_result)

        if not result.device_grant_required:
            return await self._os_game_login(result.account.uid, result.account.token)

        mmt = await self._send_game_verification_email(result.account.device_grant_ticket)
        if mmt:
            if geetest_solver:
                mmt_result = await geetest_solver(mmt)
            else:
                mmt_result = await server.solve_geetest(mmt, port=port)

            await self._send_game_verification_email(result.account.device_grant_ticket, mmt_result=mmt_result)

        code = await server.enter_code()
        verification_result = await self._verify_game_email(code, result.account.device_grant_ticket)

        return await self._os_game_login(result.account.uid, verification_result.game_token)



"""Starrail base character model."""

from genshin.models.model import APIModel, Unique


class StarRailBaseCharacter(APIModel, Unique):
    """Base character model."""

    id: int
    element: str
    rarity: int
    icon: str


class StarRailPartialCharacter(StarRailBaseCharacter):
    """Character without any equipment."""

    name: str
    level: int
    rank: int


class FloorCharacter(StarRailBaseCharacter):
    """Character in a floor."""

    level: int
    rank: int


class RogueCharacter(StarRailBaseCharacter):
    """Rogue character model."""

    level: int
    rank: int



"""Constants hardcoded for optimizations."""

from . import types

__all__ = ["LANGS"]


LANGS = {
    "zh-cn": "简体中文",
    "zh-tw": "繁體中文",
    "de-de": "Deutsch",
    "en-us": "English",
    "es-es": "Español",
    "fr-fr": "Français",
    "id-id": "Indonesia",
    "it-it": "Italiano",
    "ja-jp": "日本語",
    "ko-kr": "한국어",
    "pt-pt": "Português",
    "ru-ru": "Pусский",
    "th-th": "ภาษาไทย",
    "vi-vn": "Tiếng Việt",
    "tr-tr": "Türkçe",
}
"""Languages supported by the API."""

DS_SALT = {
    types.Region.OVERSEAS: "6s25p5ox5y14umn1p61aqyyvbvvl3lrt",
    types.Region.CHINESE: "xV8v4Qu54lUKrEYFZkJhB8cuOh9Asafs",
    "app_login": "IZPgfb0dRPtBeLuFkdDznSZ6f4wWt6y2",
    "cn_signin": "LyD1rXqMv2GJhnwdvCBjFOKGiKuLY3aO",
    "cn_passport": "JwYDpKvLj6MrMqqYU6jTKF17KNO2PXoS",
}
"""Dynamic Secret Salts."""

MIYOUSHE_GEETEST_RETCODES = {10035, 5003, 10041, 1034}
"""API error codes that indicate a Geetest was triggered during this Miyoushe API request."""

APP_KEYS = {
    types.Game.GENSHIN: {
        types.Region.OVERSEAS: "6a4c78fe0356ba4673b8071127b28123",
        types.Region.CHINESE: "d0d3a7342df2026a70f650b907800111",
    },
    types.Game.STARRAIL: {
        types.Region.OVERSEAS: "d74818dabd4182d4fbac7f8df1622648",
        types.Region.CHINESE: "4650f3a396d34d576c3d65df26415394",
    },
    types.Game.HONKAI: {
        types.Region.OVERSEAS: "243187699ab762b682a2a2e50ba02285",
        types.Region.CHINESE: "0ebc517adb1b62c6b408df153331f9aa",
    },
    types.Game.ZZZ: {
        types.Region.OVERSEAS: "ff0f2776bf515d79d1f8ff1fb98b2a06",
        types.Region.CHINESE: "4650f3a396d34d576c3d65df26415394",
    },
}
"""App keys used for game login."""

APP_IDS = {
    types.Game.GENSHIN: {
        types.Region.OVERSEAS: "4",
        types.Region.CHINESE: "4",
    },
    types.Game.STARRAIL: {
        types.Region.OVERSEAS: "11",
        types.Region.CHINESE: "8",
    },
    types.Game.HONKAI: {
        types.Region.OVERSEAS: "8",
        types.Region.CHINESE: "1",
    },
    types.Game.ZZZ: {
        types.Region.OVERSEAS: "15",
        types.Region.CHINESE: "12",
    },
}
"""App IDs used for game login."""



"""Genshin wish models."""

import enum
import json
import typing
import unicodedata

if typing.TYPE_CHECKING:
    import pydantic.v1 as pydantic
else:
    try:
        import pydantic.v1 as pydantic
    except ImportError:
        import pydantic

from genshin.models.model import Aliased, APIModel, Unique

__all__ = [
    "ArtifactPreview",
    "CharacterPreview",
    "EnemyPreview",
    "WeaponPreview",
    "WikiPage",
    "WikiPageType",
]


class WikiPageType(enum.IntEnum):
    """Wiki page types."""

    UNKNOWN = 0
    CHARACTER = 2
    WEAPON = 4
    ARTIFACT = 5
    ENEMY = 7


class BaseWikiPreview(APIModel, Unique):
    """Base wiki preview."""

    id: int = Aliased("entry_page_id")
    icon: str = Aliased("icon_url")
    name: str

    @pydantic.root_validator(pre=True)
    def __unpack_filter_values(cls, values: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
        filter_values = {
            key.split("_", 1)[1]: value["values"][0]
            for key, value in values.get("filter_values", {}).items()
            if value["values"]
        }
        values.update(filter_values)
        return values

    @pydantic.root_validator(pre=True)
    def __flatten_display_field(cls, values: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
        values.update(values.get("display_field", {}))
        return values


# shuffle validators around because of nesting
BaseWikiPreview.__pre_root_validators__.reverse()


class CharacterPreview(BaseWikiPreview):
    """Character wiki preview."""

    bonus_attribute: str = Aliased("", "property")
    rarity: int
    region: str = ""
    element: str = Aliased("vision", "")
    weapon: str

    @pydantic.validator("rarity", pre=True)
    def __extract_rarity(cls, value: typing.Union[int, str]) -> int:
        if not isinstance(value, str):
            return value

        if value[0].isdigit():
            return int(value[0])

        return int(unicodedata.numeric(value[0]))


class WeaponPreview(BaseWikiPreview):
    """Weapon wiki preview."""

    bonus_attribute: str = Aliased("property")
    rarity: int
    type: str

    @pydantic.validator("rarity", pre=True)
    def __extract_rarity(cls, value: typing.Union[int, str]) -> int:
        if not isinstance(value, str):
            return value

        if value[0].isdigit():
            return int(value[0])

        return int(unicodedata.numeric(value[0]))


class ArtifactPreview(BaseWikiPreview):
    """Artifact wiki preview."""

    effect: str

    circlet_icon: str = Aliased("circlet_of_logos_icon_url")
    flower_icon: str = Aliased("flower_of_life_icon_url")
    goblet_icon: str = Aliased("goblet_of_eonothem_icon_url")
    plume_icon: str = Aliased("plume_of_death_icon_url")
    sands_icon: str = Aliased("sands_of_eon_icon_url")

    effects: typing.Mapping[int, str]

    @pydantic.root_validator(pre=True)
    def __group_effects(cls, values: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
        effects = {
            1: values["single_set_effect"],
            2: values["two_set_effect"],
            4: values["four_set_effect"],
        }
        values["effects"] = {amount: effect for amount, effect in effects.items() if effect}
        return values


class EnemyPreview(BaseWikiPreview):
    """Enemy wiki preview."""

    drop_materials: typing.Sequence[str]

    @pydantic.validator("drop_materials", pre=True)
    def __parse_drop_materials(cls, value: typing.Union[str, typing.Sequence[str]]) -> typing.Sequence[str]:
        return json.loads(value) if isinstance(value, str) else value


_ENTRY_PAGE_MODELS: typing.Mapping[WikiPageType, typing.Type[BaseWikiPreview]] = {
    WikiPageType.CHARACTER: CharacterPreview,
    WikiPageType.WEAPON: WeaponPreview,
    WikiPageType.ARTIFACT: ArtifactPreview,
    WikiPageType.ENEMY: EnemyPreview,
}


class WikiPage(APIModel):
    """Wiki page."""

    id: int
    page_type: WikiPageType = Aliased("menu_id")

    description: str = Aliased("desc")
    header: str = Aliased("header_img_url")
    icon: str = Aliased("icon_url")

    modules: typing.Mapping[str, typing.Mapping[str, typing.Any]]

    @pydantic.validator("modules", pre=True)
    def __format_modules(
        cls,
        value: typing.Union[typing.List[typing.Dict[str, typing.Any]], typing.Dict[str, typing.Any]],
    ) -> typing.Dict[str, typing.Any]:
        if isinstance(value, typing.Mapping):
            return value

        modules: typing.Dict[str, typing.Dict[str, typing.Any]] = {}
        for module in value:
            components: typing.Dict[str, typing.Dict[str, typing.Any]] = {
                component["component_id"]: json.loads(component["data"] or "{}") for component in module["components"]
            }

            components.pop("map", None)  # not worth storing
            if "reliquary_set_effect" in components:
                # Attributes of artifacts should be flattened
                components["baseInfo"]["reliquary_set_effect"] = components.pop("reliquary_set_effect")

            _, modules[module["name"]] = components.popitem()
            continue

        return modules



"""Types used in the library."""

import enum
import typing

if typing.TYPE_CHECKING:
    from genshin.models.model import Unique

__all__ = ["Game", "Region"]

UniqueT = typing.TypeVar("UniqueT", bound="Unique")


class Region(str, enum.Enum):
    """Region to get data from."""

    OVERSEAS = "os"
    """Applies to all overseas APIs."""

    CHINESE = "cn"
    """Applies to all chinese mainland APIs."""


class Game(str, enum.Enum):
    """Hoyoverse game."""

    GENSHIN = "genshin"
    """Genshin Impact"""

    HONKAI = "honkai3rd"
    """Honkai Impact 3rd"""

    STARRAIL = "hkrpg"
    """Honkai Star Rail"""

    ZZZ = "nap"
    """Zenless Zone Zero"""


IDOr = typing.Union[int, UniqueT]
"""Allows partial objects."""



"""Private and confidential models.

All models will lack private data by design.
"""

from genshin.models.model import APIModel

__all__ = ["AccountInfo"]


class AccountInfo(APIModel):
    """Account info."""

    account_id: int
    account_name: str
    weblogin_token: str

    @property
    def login_ticket(self) -> str:
        return self.weblogin_token



"""Wish component."""

import asyncio
import functools
import typing
import urllib.parse
import warnings

from genshin import paginators, types, utility
from genshin.client import cache as client_cache
from genshin.client import routes
from genshin.client.components import base
from genshin.models.genshin import gacha as models
from genshin.utility import deprecation

__all__ = ["WishClient"]


class WishClient(base.BaseClient):
    """Wish component."""

    async def request_gacha_info(
        self,
        endpoint: str,
        *,
        lang: typing.Optional[str] = None,
        game: typing.Optional[types.Game] = None,
        authkey: typing.Optional[str] = None,
        params: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        **kwargs: typing.Any,
    ) -> typing.Mapping[str, typing.Any]:
        """Make a request towards the gacha info endpoint."""
        params = dict(params or {})
        game = game or self.default_game
        if game is None:
            raise RuntimeError("No game provided")

        authkey = authkey or self.authkeys.get(game)
        if authkey is None:
            raise RuntimeError("No authkey provided")

        base_url = routes.GACHA_URL.get_url(self.region, game)
        url = base_url / endpoint

        params["authkey_ver"] = 1
        params["authkey"] = urllib.parse.unquote(authkey)
        params["lang"] = utility.create_short_lang_code(lang or self.lang)
        params["game_biz"] = utility.get_prod_game_biz(self.region, game)

        return await self.request(url, params=params, **kwargs)

    async def _get_gacha_page(
        self,
        end_id: int,
        banner_type: int,
        *,
        game: typing.Optional[types.Game] = None,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
    ) -> typing.Sequence[typing.Any]:
        """Get a single page of wishes."""
        data = await self.request_gacha_info(
            "getGachaLog",
            lang=lang,
            game=game,
            authkey=authkey,
            params=dict(gacha_type=banner_type, size=20, end_id=end_id),
        )
        return data["list"]

    async def _get_wish_page(
        self,
        end_id: int,
        banner_type: models.GenshinBannerType,
        *,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
    ) -> typing.Sequence[models.Wish]:
        """Get a single page of wishes."""
        data = await self._get_gacha_page(
            end_id=end_id,
            banner_type=banner_type,
            lang=lang,
            authkey=authkey,
            game=types.Game.GENSHIN,
        )

        banner_names = await self.get_banner_names(lang=lang, authkey=authkey)
        return [models.Wish(**i, banner_name=banner_names[banner_type]) for i in data]

    async def _get_warp_page(
        self,
        end_id: int,
        banner_type: models.StarRailBannerType,
        *,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
    ) -> typing.Sequence[models.Warp]:
        """Get a single page of warps."""
        data = await self._get_gacha_page(
            end_id=end_id,
            banner_type=banner_type,
            lang=lang,
            authkey=authkey,
            game=types.Game.STARRAIL,
        )

        return [models.Warp(**i, banner_type=banner_type) for i in data]

    def wish_history(
        self,
        banner_type: typing.Optional[typing.Union[int, typing.Sequence[int]]] = None,
        *,
        limit: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
        end_id: int = 0,
    ) -> paginators.Paginator[models.Wish]:
        """Get the wish history of a user."""
        banner_types = banner_type or [100, 200, 301, 302]

        if not isinstance(banner_types, typing.Sequence):
            banner_types = [banner_types]

        iterators: typing.List[paginators.Paginator[models.Wish]] = []
        for banner in banner_types:
            iterators.append(
                paginators.CursorPaginator(
                    functools.partial(
                        self._get_wish_page,
                        banner_type=typing.cast(models.GenshinBannerType, banner),
                        lang=lang,
                        authkey=authkey,
                    ),
                    limit=limit,
                    end_id=end_id,
                )
            )

        if len(iterators) == 1:
            return iterators[0]

        return paginators.MergedPaginator(iterators, key=lambda wish: wish.time.timestamp())

    def warp_history(
        self,
        banner_type: typing.Optional[typing.Union[int, typing.Sequence[int]]] = None,
        *,
        limit: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
        end_id: int = 0,
    ) -> paginators.Paginator[models.Warp]:
        """Get the warp history of a user."""
        banner_types = banner_type or [1, 2, 11, 12]

        if not isinstance(banner_types, typing.Sequence):
            banner_types = [banner_types]

        iterators: typing.List[paginators.Paginator[models.Warp]] = []
        for banner in banner_types:
            iterators.append(
                paginators.CursorPaginator(
                    functools.partial(
                        self._get_warp_page,
                        banner_type=typing.cast(models.StarRailBannerType, banner),
                        lang=lang,
                        authkey=authkey,
                    ),
                    limit=limit,
                    end_id=end_id,
                )
            )

        if len(iterators) == 1:
            return iterators[0]

        return paginators.MergedPaginator(iterators, key=lambda wish: wish.time.timestamp())

    @deprecation.deprecated("get_genshin_banner_names")
    async def get_banner_names(
        self,
        *,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
    ) -> typing.Mapping[int, str]:
        """Get a list of banner names."""
        return await self.get_genshin_banner_names(lang=lang, authkey=authkey)

    async def get_genshin_banner_names(
        self,
        *,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
    ) -> typing.Mapping[int, str]:
        """Get a list of banner names."""
        data = await self.request_gacha_info(
            "getConfigList",
            lang=lang,
            game=types.Game.GENSHIN,
            authkey=authkey,
            static_cache=client_cache.cache_key("banner", endpoint="names", lang=lang or self.lang),
        )
        return {int(i["key"]): i["name"] for i in data["gacha_type_list"]}

    async def _get_banner_details(
        self,
        banner_id: str,
        *,
        game: typing.Optional[types.Game] = None,
        lang: typing.Optional[str] = None,
    ) -> models.BannerDetails:
        """Get details of a specific banner using its id."""
        lang = lang or self.lang
        game = game or self.default_game
        if game is None:
            raise RuntimeError("No game provided")
        if game == types.Game.STARRAIL:
            warnings.warn("Banner details for Star Rail are not fully supported.")

        region = "hkrpg" if game == types.Game.STARRAIL else "hk4e"
        server = "prod_official_asia" if game == types.Game.STARRAIL else "os_asia"

        data = await self.request_webstatic(
            f"/gacha_info/{region}/{server}/{banner_id}/{lang}.json",
            cache=client_cache.cache_key("banner", endpoint="details", banner=banner_id, lang=lang),
        )
        return models.BannerDetails(**data, banner_id=banner_id)

    @deprecation.deprecated("get_genshin_banner_ids")
    async def get_banner_ids(self) -> typing.Sequence[str]:
        """Get a list of banner ids.

        Uses the current cn banners.
        """
        return await self.get_genshin_banner_ids()

    async def get_genshin_banner_ids(self) -> typing.Sequence[str]:
        """Get a list of banner ids.

        Uses the current cn banners.
        """

        def process_gacha(data: typing.Mapping[str, typing.Any]) -> str:
            # Temporary fix for 4.5 chronicled wish
            if data["gacha_type"] == 500:
                return "8b10b48c52dd6870f92d72e9963b44bb8968ed2f"
            return data["gacha_id"]

        data = await self.request_webstatic(
            "gacha_info/hk4e/cn_gf01/gacha/list.json",
            region=types.Region.CHINESE,
            cache=client_cache.cache_key("banner", endpoint="ids"),
        )
        return list(map(process_gacha, data["data"]["list"]))

    async def get_banner_details(
        self,
        banner_ids: typing.Optional[typing.Sequence[str]] = None,
        *,
        game: typing.Optional[types.Game] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.BannerDetails]:
        """Get all banner details at once in a batch."""
        game = game or self.default_game
        if game == types.Game.STARRAIL and not banner_ids:
            raise RuntimeError("No banner ids provided for star rail")

        banner_ids = banner_ids or await self.get_genshin_banner_ids()

        coros = (self._get_banner_details(i, lang=lang, game=game) for i in banner_ids)
        data = await asyncio.gather(*coros)
        return list(data)

    @deprecation.deprecated("get_genshin_gacha_items")
    async def get_gacha_items(
        self,
        *,
        server: str = "os_asia",
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.GachaItem]:
        """Get the list of characters and weapons that can be gotten from the gacha."""
        return await self.get_genshin_gacha_items(server=server, lang=lang)

    async def get_genshin_gacha_items(
        self,
        *,
        server: str = "os_asia",
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.GachaItem]:
        """Get the list of characters and weapons that can be gotten from the gacha."""
        lang = lang or self.lang
        data = await self.request_webstatic(
            f"/hk4e/gacha_info/{server}/items/{lang}.json",
            cache=client_cache.cache_key("banner", endpoint="items", lang=lang),
        )
        return [models.GachaItem(**i) for i in data]



"""Genshin wish models."""

import datetime
import enum
import re
import typing

if typing.TYPE_CHECKING:
    import pydantic.v1 as pydantic
else:
    try:
        import pydantic.v1 as pydantic
    except ImportError:
        import pydantic

from genshin.models.model import Aliased, APIModel, Unique

__all__ = [
    "BannerDetailItem",
    "BannerDetails",
    "BannerDetailsUpItem",
    "GachaItem",
    "GenshinBannerType",
    "Warp",
    "Wish",
]


class GenshinBannerType(enum.IntEnum):
    """Banner types in wish histories."""

    NOVICE = 100
    """Temporary novice banner."""

    STANDARD = PERMANENT = 200
    """Permanent standard banner."""

    CHARACTER = 301
    """Rotating character banner."""

    WEAPON = 302
    """Rotating weapon banner."""

    # these are special cases
    # they exist inside the history but should be counted as the same

    CHARACTER1 = 301
    """Character banner #1."""

    CHARACTER2 = 400
    """Character banner #2."""


class StarRailBannerType(enum.IntEnum):
    """Banner types in wish histories."""

    STANDARD = PERMANENT = 1
    """Permanent standard banner."""
    NOVICE = 2
    """Temporary novice banner."""
    CHARACTER = 11
    """Rotating character banner."""
    WEAPON = 12
    """Rotating weapon banner."""


class Wish(APIModel, Unique):
    """Wish made on any banner."""

    uid: int

    id: int
    type: str = Aliased("item_type")
    name: str
    rarity: int = Aliased("rank_type")
    time: datetime.datetime

    banner_type: GenshinBannerType = Aliased("gacha_type")
    banner_name: str

    @pydantic.validator("banner_type", pre=True)
    def __cast_banner_type(cls, v: typing.Any) -> int:
        return int(v)


class Warp(APIModel, Unique):
    """Warp made on any banner."""

    uid: int

    id: int
    item_id: int
    type: str = Aliased("item_type")
    name: str
    rarity: int = Aliased("rank_type")
    time: datetime.datetime

    banner_type: StarRailBannerType
    banner_id: int = Aliased("gacha_id")

    @pydantic.validator("banner_type", pre=True)
    def __cast_banner_type(cls, v: typing.Any) -> int:
        return int(v)


class BannerDetailItem(APIModel):
    """Item that may be gotten from a banner."""

    name: str = Aliased("item_name")
    type: str = Aliased("item_type")
    rarity: int = Aliased("rank")

    up: bool = Aliased("is_up")
    order: int = Aliased("order_value")


class BannerDetailsUpItem(APIModel):
    """Item that has a rate-up on a banner."""

    name: str = Aliased("item_name")
    type: str = Aliased("item_type")
    element: str = Aliased("item_attr")
    icon: str = Aliased("item_img")

    @pydantic.validator("element", pre=True)
    def __parse_element(cls, v: str) -> str:
        return {
            "风": "Anemo",
            "火": "Pyro",
            "水": "Hydro",
            "雷": "Electro",
            "冰": "Cryo",
            "岩": "Geo",
            "草": "Dendro",
            "": "",
        }.get(v, v)


class BannerDetails(APIModel):
    """Details of a banner."""

    banner_id: str
    banner_type: int = Aliased("gacha_type")
    title: str
    content: str
    date_range: str

    r5_up_prob: typing.Optional[float]
    r4_up_prob: typing.Optional[float]
    r5_prob: typing.Optional[float]
    r4_prob: typing.Optional[float]
    r3_prob: typing.Optional[float]
    r5_guarantee_prob: typing.Optional[float] = Aliased("r5_baodi_prob")
    r4_guarantee_prob: typing.Optional[float] = Aliased("r4_baodi_prob")
    r3_guarantee_prob: typing.Optional[float] = Aliased("r3_baodi_prob")

    r5_up_items: typing.Sequence[BannerDetailsUpItem]
    r4_up_items: typing.Sequence[BannerDetailsUpItem]

    r5_items: typing.List[BannerDetailItem] = Aliased("r5_prob_list")
    r4_items: typing.List[BannerDetailItem] = Aliased("r4_prob_list")
    r3_items: typing.List[BannerDetailItem] = Aliased("r3_prob_list")

    @pydantic.validator("r5_up_items", "r4_up_items", pre=True)
    def __replace_none(cls, v: typing.Optional[typing.Sequence[typing.Any]]) -> typing.Sequence[typing.Any]:
        return v or []

    @pydantic.validator(
        "r5_up_prob",
        "r4_up_prob",
        "r5_prob",
        "r4_prob",
        "r3_prob",
        "r5_guarantee_prob",
        "r4_guarantee_prob",
        "r3_guarantee_prob",
        pre=True,
    )
    def __parse_percentage(cls, v: typing.Optional[str]) -> typing.Optional[float]:
        if v is None or isinstance(v, (int, float)):
            return v

        return None if v == "0%" else float(v[:-1].replace(",", "."))

    @property
    def name(self) -> str:
        return re.sub(r"<.*?>", "", self.title).strip()

    @property
    def banner_type_name(self) -> str:
        banners = {
            100: "Novice Wishes",
            200: "Permanent Wish",
            301: "Character Event Wish",
            302: "Weapon Event Wish",
            400: "Character Event Wish",
        }
        return banners[self.banner_type]

    @property
    def items(self) -> typing.Sequence[BannerDetailItem]:
        items = self.r5_items + self.r4_items + self.r3_items
        return sorted(items, key=lambda x: x.order)


class GachaItem(APIModel, Unique):
    """Item that can be gotten from the gacha."""

    name: str
    type: str = Aliased("item_type")
    rarity: int = Aliased("rank_type")
    id: int = Aliased("item_id")

    @pydantic.validator("id")
    def __format_id(cls, v: int) -> int:
        return 10000000 + v - 1000 if len(str(v)) == 4 else v

    def is_character(self) -> bool:
        """Whether this item is a character."""
        return len(str(self.id)) == 8



"""Geetest-related models"""

import enum
import json
import typing

if typing.TYPE_CHECKING:
    import pydantic.v1 as pydantic
else:
    try:
        import pydantic.v1 as pydantic
    except ImportError:
        import pydantic

from genshin.utility import auth as auth_utility

__all__ = [
    "MMT",
    "BaseMMT",
    "BaseMMTResult",
    "BaseSessionMMTResult",
    "MMTResult",
    "MMTv4",
    "MMTv4Result",
    "RiskyCheckMMT",
    "RiskyCheckMMTResult",
    "SessionMMT",
    "SessionMMTResult",
    "SessionMMTv4",
    "SessionMMTv4Result",
]


class BaseMMT(pydantic.BaseModel):
    """Base Geetest verification data model."""

    new_captcha: int
    success: int

    @pydantic.root_validator(pre=True)
    def __parse_data(cls, data: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
        """Parse the data if it was provided in a raw format."""
        if "data" in data:
            # Assume the data is aigis header and parse it
            session_id = data["session_id"]
            data = data["data"]
            if isinstance(data, str):
                data = json.loads(data)

            data["session_id"] = session_id

        return data


class MMT(BaseMMT):
    """Geetest verification data."""

    challenge: str
    gt: str


class SessionMMT(MMT):
    """Session-based geetest verification data."""

    session_id: str

    def get_mmt(self) -> MMT:
        """Get the base MMT data."""
        return MMT(**self.dict(exclude={"session_id"}))


class MMTv4(BaseMMT):
    """Geetest verification data (V4)."""

    captcha_id: str = pydantic.Field(alias="gt")
    risk_type: str


class SessionMMTv4(MMTv4):
    """Session-based geetest verification data (V4)."""

    session_id: str

    def get_mmt(self) -> MMTv4:
        """Get the base MMTv4 data."""
        return MMTv4(**self.dict(exclude={"session_id"}))


class RiskyCheckMMT(MMT):
    """MMT returned by the risky check endpoint."""

    check_id: str


class BaseMMTResult(pydantic.BaseModel):
    """Base Geetest verification result model."""

    def get_data(self) -> typing.Dict[str, typing.Any]:
        """Get the base MMT result data.

        This method acts as `dict` but excludes the `session_id` field.
        """
        return self.dict(exclude={"session_id"})


class BaseSessionMMTResult(BaseMMTResult):
    """Base session-based Geetest verification result model."""

    session_id: str

    def to_aigis_header(self) -> str:
        """Convert the result to `x-rpc-aigis` header."""
        return auth_utility.get_aigis_header(self.session_id, self.get_data())


class MMTResult(BaseMMTResult):
    """Geetest verification result."""

    geetest_challenge: str
    geetest_validate: str
    geetest_seccode: str


class SessionMMTResult(MMTResult, BaseSessionMMTResult):
    """Session-based geetest verification result."""


class MMTv4Result(BaseMMTResult):
    """Geetest verification result (V4)."""

    captcha_id: str
    lot_number: str
    pass_token: str
    gen_time: str
    captcha_output: str


class SessionMMTv4Result(MMTv4Result, BaseSessionMMTResult):
    """Session-based geetest verification result (V4)."""


class RiskyCheckMMTResult(MMTResult):
    """Risky check MMT result."""

    check_id: str

    def to_rpc_risky(self) -> str:
        """Convert the MMT result to a RPC risky header."""
        return auth_utility.generate_risky_header(self.check_id, self.geetest_challenge, self.geetest_validate)


class RiskyCheckAction(str, enum.Enum):
    """Risky check action returned by the API."""

    ACTION_NONE = "ACTION_NONE"
    """No action required."""
    ACTION_GEETEST = "ACTION_GEETEST"
    """Geetest verification required."""


class RiskyCheckResult(pydantic.BaseModel):
    """Model for the risky check result."""

    id: str
    action: RiskyCheckAction
    mmt: typing.Optional[MMT] = pydantic.Field(alias="geetest")

    def to_mmt(self) -> RiskyCheckMMT:
        """Convert the check result to a `RiskyCheckMMT` object."""
        if self.mmt is None:
            raise ValueError("The check result does not contain a MMT object.")

        return RiskyCheckMMT(**self.mmt.dict(), check_id=self.id)



"""Honkai stats models."""

import typing

if typing.TYPE_CHECKING:
    import pydantic.v1 as pydantic
else:
    try:
        import pydantic.v1 as pydantic
    except ImportError:
        import pydantic

from genshin.models import hoyolab
from genshin.models.model import Aliased, APIModel

from . import battlesuits as honkai_battlesuits
from . import modes

__all__ = [
    "FullHonkaiUserStats",
    "HonkaiStats",
    "HonkaiUserStats",
]


def _model_to_dict(model: APIModel, lang: str = "en-us") -> typing.Mapping[str, typing.Any]:
    """Turn fields into properly named ones."""
    ret: typing.Dict[str, typing.Any] = {}
    for field in model.__fields__.values():
        if not field.field_info.extra.get("mi18n"):
            continue

        mi18n = model._get_mi18n(field, lang)
        val = getattr(model, field.name)
        if isinstance(val, APIModel):
            ret[mi18n] = _model_to_dict(val, lang)
        else:
            ret[mi18n] = val

    return ret


# flake8: noqa: E222
class MemorialArenaStats(APIModel):
    """Represents a user's stats regarding the Memorial Arena gamemodes."""

    # fmt: off
    ranking: float = Aliased("battle_field_ranking_percentage", mi18n="bbs/battle_field_ranking_percentage")
    raw_rank: int =  Aliased("battle_field_rank",               mi18n="bbs/rank")
    score: int =     Aliased("battle_field_score",              mi18n="bbs/score")
    raw_tier: int =  Aliased("battle_field_area",               mi18n="bbs/settled_level")
    # fmt: on

    @pydantic.validator("ranking", pre=True)
    def __normalize_ranking(cls, value: typing.Union[str, float]) -> float:
        return float(value) if value else 0

    def as_dict(self, lang: typing.Optional[str] = None) -> typing.Mapping[str, typing.Any]:
        return _model_to_dict(self, lang or self.lang)

    @property
    def rank(self) -> str:
        """The user's Memorial Arena rank as displayed in-game."""
        return modes.prettify_MA_rank(self.raw_rank)

    @property
    def tier(self) -> str:
        """The user's Memorial Arena tier as displayed in-game."""
        return self.get_tier()

    def get_tier(self, lang: typing.Optional[str] = None) -> str:
        """Get the user's Memorial Arena tier in a specific language."""
        key = modes.get_competitive_tier_mi18n(self.raw_tier)
        return self._get_mi18n(key, lang or self.lang)


# flake8: noqa: E222
class SuperstringAbyssStats(APIModel):
    """Represents a user's stats regarding Superstring Abyss."""

    # fmt: off
    raw_rank: int = Aliased("level",             mi18n="bbs/rank")
    trophies: int = Aliased("cup_number",        mi18n="bbs/cup_number")
    score: int =    Aliased("abyss_score",       mi18n="bbs/explain_text_2")
    raw_tier: int = Aliased("battle_field_area", mi18n="bbs/settled_level")
    # fmt: on

    # for consistency between types; also allows us to forego the mi18n fuckery
    latest_type: typing.ClassVar[str] = "Superstring"

    def as_dict(self, lang: typing.Optional[str] = None) -> typing.Mapping[str, typing.Any]:
        return _model_to_dict(self, lang or self.lang)

    @property
    def rank(self) -> str:
        """The user's Abyss rank as displayed in-game."""
        return self.get_rank()

    def get_rank(self, lang: typing.Optional[str] = None) -> str:
        """Get the user's Abyss rank in a specific language."""
        key = modes.get_abyss_rank_mi18n(self.raw_rank, self.raw_tier)
        return self._get_mi18n(key, lang or self.lang)

    @property
    def tier(self) -> str:
        """The user's Abyss tier as displayed in-game."""
        return self.get_tier()

    def get_tier(self, lang: typing.Optional[str] = None) -> str:
        """Get the user's Abyss tier in a specific language."""
        key = modes.get_competitive_tier_mi18n(self.raw_tier)
        return self._get_mi18n(key, lang or self.lang)


# flake8: noqa: E222
class OldAbyssStats(APIModel):
    """Represents a user's stats regarding Q-Singularis and Dirac Sea."""

    # fmt: off
    raw_q_singularis_rank: typing.Optional[int] = Aliased("level_of_quantum", mi18n="bbs/Quantum")
    raw_dirac_sea_rank: typing.Optional[int] =    Aliased("level_of_ow",      mi18n="bbs/level_of_ow")
    score: int =                                  Aliased("abyss_score",      mi18n="bbs/explain_text_2")
    raw_tier: int =                               Aliased("latest_area",      mi18n="bbs/settled_level")
    raw_latest_rank: typing.Optional[int] =       Aliased("latest_level",     mi18n="bbs/rank")
    # TODO: Add proper key
    latest_type: str =                            Aliased(                    mi18n="bbs/latest_type")
    # fmt: on

    @pydantic.validator("raw_q_singularis_rank", "raw_dirac_sea_rank", "raw_latest_rank", pre=True)
    def __normalize_rank(cls, rank: typing.Optional[str]) -> typing.Optional[int]:  # modes.OldAbyss.__normalize_rank
        if isinstance(rank, int):
            return rank

        if rank is None or "Unknown" in rank:
            return None

        return 69 - ord(rank)

    @property
    def q_singularis_rank(self) -> typing.Optional[str]:
        """The user's latest Q-Singularis rank as displayed in-game."""
        if self.raw_q_singularis_rank is None:
            return None

        return self.get_rank(self.raw_q_singularis_rank)

    @property
    def dirac_sea_rank(self) -> typing.Optional[str]:
        """The user's latest Dirac Sea rank as displayed in-game."""
        if self.raw_dirac_sea_rank is None:
            return None

        return self.get_rank(self.raw_dirac_sea_rank)

    @property
    def latest_rank(self) -> typing.Optional[str]:
        """The user's latest Abyss rank as displayed in-game. Seems to apply after weekly reset,
        so this may differ from the user's Dirac Sea/Q-Singularis ranks if their rank changed.
        """
        if self.raw_latest_rank is None:
            return None

        return self.get_rank(self.raw_latest_rank)

    def get_rank(self, rank: int, lang: typing.Optional[str] = None) -> str:
        """Get the user's Abyss rank in a specific language.

        Must be supplied with one of the raw ranks stored on this class.
        """
        key = modes.get_abyss_rank_mi18n(rank, self.raw_tier)
        return self._get_mi18n(key, lang or self.lang)

    @property
    def tier(self) -> str:
        """The user's Abyss tier as displayed in-game."""
        return modes.get_competitive_tier_mi18n(self.raw_tier)

    def get_tier(self, lang: typing.Optional[str] = None) -> str:
        """Get the user's Abyss tier in a specific language."""
        key = modes.get_competitive_tier_mi18n(self.raw_tier)
        return self._get_mi18n(key, lang or self.lang)

    def as_dict(self, lang: typing.Optional[str] = None) -> typing.Mapping[str, typing.Any]:
        return _model_to_dict(self, lang or self.lang)

    class Config:
        # this is for the "stat_lang" field, hopefully nobody abuses this
        allow_mutation = True


# flake8: noqa: E222
class ElysianRealmStats(APIModel):
    """Represents a user's stats regarding Elysian Realms."""

    # fmt: off
    highest_difficulty: int = Aliased("god_war_max_punish_level",        mi18n="bbs/god_war_max_punish_level")
    remembrance_sigils: int = Aliased("god_war_extra_item_number",       mi18n="bbs/god_war_extra_item_number")
    highest_score: int =      Aliased("god_war_max_challenge_score",     mi18n="bbs/god_war_max_challenge_score")
    highest_floor: int =      Aliased("god_war_max_challenge_level",     mi18n="bbs/rogue_setted_level")
    max_level_suits: int =    Aliased("god_war_max_level_avatar_number", mi18n="bbs/explain_text_6")
    # fmt: on

    def as_dict(self, lang: str = "en-us") -> typing.Mapping[str, typing.Any]:
        return _model_to_dict(self, lang)


class HonkaiStats(APIModel):
    """Represents a user's stat page"""

    # fmt: off
    active_days: int =     Aliased("active_day_number",         mi18n="bbs/active_day_number")
    achievements: int =    Aliased("achievement_number",        mi18n="bbs/achievment_complete_count")

    battlesuits: int =     Aliased("armor_number",              mi18n="bbs/armor_number")
    battlesuits_SSS: int = Aliased("sss_armor_number",          mi18n="bbs/sss_armor_number")
    stigmata: int =        Aliased("stigmata_number",           mi18n="bbs/stigmata_number")
    stigmata_5star: int =  Aliased("five_star_stigmata_number", mi18n="bbs/stigmata_number_5")
    weapons: int =         Aliased("weapon_number",             mi18n="bbs/weapon_number")
    weapons_5star: int =   Aliased("five_star_weapon_number",   mi18n="bbs/weapon_number_5")
    outfits: int =         Aliased("suit_number",               mi18n="bbs/suit_number")
    # fmt: on

    abyss: typing.Union[SuperstringAbyssStats, OldAbyssStats] = Aliased(mi18n="bbs/explain_text_1")
    memorial_arena: MemorialArenaStats = Aliased(mi18n="bbs/battle_field_ranking_percentage")
    elysian_realm: ElysianRealmStats = Aliased(mi18n="bbs/godwor")

    @pydantic.root_validator(pre=True)
    def __pack_gamemode_stats(cls, values: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
        if "new_abyss" in values:
            values["abyss"] = SuperstringAbyssStats(**values["new_abyss"], **values)
        elif "old_abyss" in values:
            values["abyss"] = OldAbyssStats(**values["old_abyss"], **values)

        if "memorial_arena" not in values:
            values["memorial_arena"] = MemorialArenaStats(**values)

        if "elysian_realm" not in values:
            values["elysian_realm"] = ElysianRealmStats(**values)

        return values

    def as_dict(self, lang: str = "en-us") -> typing.Mapping[str, typing.Any]:
        """Turn fields into properly named ones."""
        return _model_to_dict(self, lang)


class HonkaiUserStats(APIModel):
    """Represents basic user stats, showing only generic user data and stats."""

    info: hoyolab.UserInfo = Aliased("role")
    stats: HonkaiStats


class FullHonkaiUserStats(HonkaiUserStats):
    """Represents a user's full stats, including characters, gear, and gamemode data"""

    battlesuits: typing.Sequence[honkai_battlesuits.FullBattlesuit]
    abyss: typing.Sequence[typing.Union[modes.SuperstringAbyss, modes.OldAbyss]]
    memorial_arena: typing.Sequence[modes.MemorialArena]
    elysian_realm: typing.Sequence[modes.ElysianRealm]

    @property
    def abyss_superstring(self) -> typing.Sequence[modes.SuperstringAbyss]:
        """Filter `self.abyss` to only return instances of Superstring Abyss."""
        return [entry for entry in self.abyss if isinstance(entry, modes.SuperstringAbyss)]

    @property
    def abyss_q_singularis(self) -> typing.Sequence[modes.OldAbyss]:
        """Filter `self.abyss` to only return instances of Q-Singularis."""
        return [entry for entry in self.abyss if isinstance(entry, modes.OldAbyss) and entry.type == "Q-Singularis"]

    @property
    def abyss_dirac_sea(self) -> typing.Sequence[modes.OldAbyss]:
        """Filter `self.abyss` to only return instances of Dirac Sea."""
        return [entry for entry in self.abyss if isinstance(entry, modes.OldAbyss) and entry.type == "Dirac Sea"]



import asyncio
import json
import os
import typing
import warnings

import pytest

import genshin


@pytest.fixture(scope="session")
def event_loop():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        loop = asyncio.get_event_loop()

    yield loop
    loop.close()


@pytest.fixture(scope="session")
def cookies() -> typing.Mapping[str, str]:
    if not os.environ.get("GENSHIN_COOKIES"):
        pytest.exit("No cookies set", 1)

    cookies = genshin.client.manager.parse_cookie(os.environ["GENSHIN_COOKIES"])

    return cookies


@pytest.fixture(scope="session")
def local_cookies() -> typing.Mapping[str, str]:
    try:
        cookies = genshin.utility.get_browser_cookies()
    except Exception:
        cookies = {}

    if not cookies:
        if not os.environ.get("LOCAL_GENSHIN_COOKIES"):
            return {}

        cookies = genshin.client.manager.parse_cookie(os.environ["LOCAL_GENSHIN_COOKIES"])

    return cookies


@pytest.fixture(scope="session")
def chinese_cookies() -> typing.Mapping[str, str]:
    if not os.environ.get("CHINESE_GENSHIN_COOKIES"):
        return {}

    cookies = genshin.client.manager.parse_cookie(os.environ["CHINESE_GENSHIN_COOKIES"])

    return cookies


@pytest.fixture(scope="session")
def local_chinese_cookies() -> typing.Mapping[str, str]:
    if not os.environ.get("LOCAL_CHINESE_GENSHIN_COOKIES"):
        return {}

    cookies = genshin.client.manager.parse_cookie(os.environ["LOCAL_CHINESE_GENSHIN_COOKIES"])

    return cookies


@pytest.fixture(scope="session")
async def cache():
    """Return a session that gets its contents dumped into a log file."""
    cache = genshin.Cache()
    yield cache

    cache = {str(key): value for key, (_, value) in cache.cache.items()}

    cache["CHARACTER_NAMES"] = {
        lang: [c._asdict() for c in chars.values()] for lang, chars in genshin.models.CHARACTER_NAMES.items()
    }
    cache["BATTLESUIT_IDENTIFIERS"] = genshin.models.BATTLESUIT_IDENTIFIERS

    os.makedirs(".pytest_cache", exist_ok=True)
    with open(".pytest_cache/hoyo_cache.json", "w", encoding="utf-8") as file:
        json.dump(cache, file, indent=4, ensure_ascii=False)


@pytest.fixture(scope="session")
async def client(cookies: typing.Mapping[str, str], cache: genshin.Cache):
    """Return a client with environment cookies."""
    client = genshin.Client()
    client.debug = True
    client.set_cookies(cookies)
    client.cache = cache

    return client


@pytest.fixture(scope="session")
async def lclient(local_cookies: typing.Mapping[str, str], cache: genshin.Cache):
    """Return the local client."""
    if not local_cookies:
        pytest.skip("Skipped local test")

    client = genshin.Client()
    client.debug = True
    client.default_game = genshin.Game.GENSHIN
    client.set_cookies(local_cookies)
    try:
        client.set_authkey()
    except (FileNotFoundError, ValueError):
        pass

    client.cache = cache

    return client


@pytest.fixture(scope="session")
async def authkey(lclient: genshin.GenshinClient):
    if lclient.authkey is None:
        pytest.skip("Skipped authkey test")

    try:
        await lclient.wish_history(200, limit=1)
    except genshin.AuthkeyException as e:
        pytest.skip(f"Skipped authkey test ({e.msg})")

    return lclient.authkey


@pytest.fixture(scope="session")
async def cnclient(chinese_cookies: typing.Mapping[str, str]):
    """Return the client with chinese cookies."""
    if not chinese_cookies:
        pytest.skip("Skipped chinese test")

    client = genshin.Client()
    client.region = genshin.types.Region.CHINESE
    client.debug = True
    client.set_cookies(chinese_cookies)

    return client


@pytest.fixture(scope="session")
async def lcnclient(local_chinese_cookies: typing.Mapping[str, str]):
    """Return the local client with chinese cookies."""
    if not local_chinese_cookies:
        pytest.skip("Skipped local chinese test")
        return

    client = genshin.Client()
    client.region = genshin.types.Region.CHINESE
    client.debug = True
    client.set_cookies(local_chinese_cookies)

    return client


@pytest.fixture(scope="session")
def genshin_uid():
    return 710785423


@pytest.fixture(scope="session")
def honkai_uid():
    return 200365120


@pytest.fixture(scope="session")
def hoyolab_id():
    return 8366222


@pytest.fixture(scope="session")
def genshin_cnuid():
    return 101322963


@pytest.fixture(scope="session")
def miyoushe_uid():
    return 75276539


def pytest_addoption(parser: pytest.Parser):
    parser.addoption("--cooperative", action="store_true")


def pytest_collection_modifyitems(items: typing.List[pytest.Item], config: pytest.Config):
    if config.option.cooperative:
        for item in items:
            if isinstance(item, pytest.Function) and asyncio.iscoroutinefunction(item.obj):
                item.add_marker("asyncio_cooperative")

    for index, item in enumerate(items):
        if "reserialization" in item.name:
            break
    else:
        return items

    item = items.pop(index)
    items.append(item)

    return items



"""Auth utilities."""

import base64
import hmac
import json
import typing
from hashlib import sha256

from genshin import constants

__all__ = ["encrypt_credentials", "generate_sign"]


# RSA key used for OS app/web login
LOGIN_KEY_TYPE_1 = b"""
-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA4PMS2JVMwBsOIrYWRluY
wEiFZL7Aphtm9z5Eu/anzJ09nB00uhW+ScrDWFECPwpQto/GlOJYCUwVM/raQpAj
/xvcjK5tNVzzK94mhk+j9RiQ+aWHaTXmOgurhxSp3YbwlRDvOgcq5yPiTz0+kSeK
ZJcGeJ95bvJ+hJ/UMP0Zx2qB5PElZmiKvfiNqVUk8A8oxLJdBB5eCpqWV6CUqDKQ
KSQP4sM0mZvQ1Sr4UcACVcYgYnCbTZMWhJTWkrNXqI8TMomekgny3y+d6NX/cFa6
6jozFIF4HCX5aW8bp8C8vq2tFvFbleQ/Q3CU56EWWKMrOcpmFtRmC18s9biZBVR/
8QIDAQAB
-----END PUBLIC KEY-----
"""

# RSA key used for CN app/game and game login
LOGIN_KEY_TYPE_2 = b"""
-----BEGIN PUBLIC KEY-----
MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDDvekdPMHN3AYhm/vktJT+YJr7
cI5DcsNKqdsx5DZX0gDuWFuIjzdwButrIYPNmRJ1G8ybDIF7oDW2eEpm5sMbL9zs
9ExXCdvqrn51qELbqj0XxtMTIpaCHFSI50PfPpTFV9Xt/hmyVwokoOXFlAEgCn+Q
CgGs52bFoYMtyi+xEQIDAQAB
-----END PUBLIC KEY-----
"""

WEB_LOGIN_HEADERS = {
    "x-rpc-app_id": "c9oqaq3s3gu8",
    "x-rpc-client_type": "4",
    # If not equal account.hoyolab.com It's will return retcode 1200 [Unauthorized]
    "Origin": "https://account.hoyolab.com",
    "Referer": "https://account.hoyolab.com/",
}

APP_LOGIN_HEADERS = {
    "x-rpc-app_id": "c9oqaq3s3gu8",
    "x-rpc-client_type": "2",
    # Passing "x-rpc-device_id" header will trigger email verification
    # (unless the device_id is already verified).
    #
    # For some reason, without this header, email verification is not triggered.
    # "x-rpc-device_id": "1c33337bd45c1bfs",
}

CN_LOGIN_HEADERS = {
    "x-rpc-app_id": "bll8iq97cem8",
    "x-rpc-client_type": "4",
    "x-rpc-source": "v2.webLogin",
    "x-rpc-device_id": "586f2440-856a-4243-8076-2b0a12314197",
}

EMAIL_SEND_HEADERS = {
    "x-rpc-app_id": "c9oqaq3s3gu8",
    "x-rpc-client_type": "2",
}

EMAIL_VERIFY_HEADERS = {
    "x-rpc-app_id": "c9oqaq3s3gu8",
    "x-rpc-client_type": "2",
}

CREATE_MMT_HEADERS = {
    "x-rpc-app_version": "2.60.1",
    "x-rpc-client_type": "5",
}

DEVICE_ID = "D6AF5103-D297-4A01-B86A-87F87DS5723E"

RISKY_CHECK_HEADERS = {
    "x-rpc-client_type": "1",
    "x-rpc-channel_id": "1",
    "x-rpc-game_biz": "hkrpg_global",
}

SHIELD_LOGIN_HEADERS = {
    "x-rpc-client_type": "1",
    "x-rpc-channel_id": "1",
    "x-rpc-game_biz": "hkrpg_global",
    "x-rpc-device_id": DEVICE_ID,
}

GRANT_TICKET_HEADERS = {
    "x-rpc-client_type": "1",
    "x-rpc-channel_id": "1",
    "x-rpc-game_biz": "hkrpg_global",
    "x-rpc-device_id": DEVICE_ID,
    "x-rpc-language": "ru",
}

GAME_LOGIN_HEADERS = {
    "x-rpc-client_type": "1",
    "x-rpc-channel_id": "1",
    "x-rpc-game_biz": "hkrpg_global",
    "x-rpc-device_id": DEVICE_ID,
}

GEETEST_LANGS = {
    "简体中文": "zh-cn",
    "繁體中文": "zh-tw",
    "Deutsch": "de",
    "English": "en",
    "Español": "es",
    "Français": "fr",
    "Indonesia": "id",
    "Italiano": "it",
    "日本語": "ja",
    "한국어": "ko",
    "Português": "pt-pt",
    "Pусский": "ru",
    "ภาษาไทย": "th",
    "Tiếng Việt": "vi",
    "Türkçe": "tr",
}


def lang_to_geetest_lang(lang: str) -> str:
    """Convert `client.lang` to geetest lang."""
    return GEETEST_LANGS.get(constants.LANGS.get(lang, "en-us"), "en")


def encrypt_credentials(text: str, key_type: typing.Literal[1, 2]) -> str:
    """Encrypt text for geetest."""
    import rsa

    public_key = rsa.PublicKey.load_pkcs1_openssl_pem(LOGIN_KEY_TYPE_1 if key_type == 1 else LOGIN_KEY_TYPE_2)
    crypto = rsa.encrypt(text.encode("utf-8"), public_key)
    return base64.b64encode(crypto).decode("utf-8")


def get_aigis_header(session_id: str, mmt_data: typing.Dict[str, typing.Any]) -> str:
    """Get aigis header."""
    return f"{session_id};{base64.b64encode(json.dumps(mmt_data).encode()).decode()}"


def generate_sign(data: typing.Dict[str, typing.Any], key: str) -> str:
    """Generate a sign for the given `data` and `app_key`."""
    string = ""
    for k in sorted(data.keys()):
        string += k + "=" + str(data[k]) + "&"
    return hmac.new(key.encode(), string[:-1].encode(), sha256).hexdigest()


def generate_risky_header(
    check_id: str,
    challenge: str = "",
    validate: str = "",
) -> str:
    """Generate risky header for geetest verification."""
    return f"id={check_id};c={challenge};s={validate}|jordan;v={validate}"



"""Public genshin announcement models."""

import datetime

from genshin.models.model import Aliased, APIModel, Unique

__all__ = ["Announcement"]


class Announcement(APIModel, Unique):
    """Announcement model."""

    id: int = Aliased("ann_id")
    title: str
    subtitle: str
    banner: str
    content: str

    type_label: str
    type: int
    tag_icon: str

    login_alert: bool
    remind: bool
    alert: bool
    remind_ver: int
    extra_remind: bool

    start_time: datetime.datetime
    end_time: datetime.datetime
    tag_start_time: datetime.datetime
    tag_end_time: datetime.datetime

    lang: str  # type: ignore
    has_content: bool



"""Wiki component."""

import typing

from genshin import types
from genshin.client import cache, routes
from genshin.client.components import base
from genshin.models.genshin import wiki as models

__all__ = ["WikiClient"]


class WikiClient(base.BaseClient):
    """Wiki component."""

    async def request_wiki(
        self,
        endpoint: str,
        *,
        lang: typing.Optional[str] = None,
        headers: typing.Optional[typing.Mapping[str, str]] = None,
        **kwargs: typing.Any,
    ) -> typing.Mapping[str, typing.Any]:
        """Make a request towards the wiki endpoint."""
        headers = dict(headers or {})

        url = routes.WIKI_URL.get_url() / endpoint
        headers["x-rpc-language"] = lang or self.lang

        return await self.request(url, headers=headers, **kwargs)

    @typing.overload
    async def get_wiki_previews(  # noqa: D102 missing docstring in overload?
        self,
        menu: typing.Literal[models.WikiPageType.CHARACTER],
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.CharacterPreview]: ...

    @typing.overload
    async def get_wiki_previews(  # noqa: D102 missing docstring in overload?
        self,
        menu: typing.Literal[models.WikiPageType.WEAPON],
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.WeaponPreview]: ...

    @typing.overload
    async def get_wiki_previews(  # noqa: D102 missing docstring in overload?
        self,
        menu: typing.Literal[models.WikiPageType.ARTIFACT],
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.ArtifactPreview]: ...

    @typing.overload
    async def get_wiki_previews(  # noqa: D102 missing docstring in overload?
        self,
        menu: typing.Literal[models.WikiPageType.ENEMY],
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.EnemyPreview]: ...

    @typing.overload
    async def get_wiki_previews(  # noqa: D102 missing docstring in overload?
        self,
        menu: int,
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.BaseWikiPreview]: ...

    async def get_wiki_previews(
        self,
        menu: int,
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.BaseWikiPreview]:
        """Get a list of wiki previews."""
        payload = dict(filters=[], menu_id=int(menu), page_num=1, page_size=1000, use_es=True)
        cache_key = cache.cache_key("wiki", endpoint="entry", menu=menu, lang=lang or self.lang)
        data = await self.request_wiki("get_entry_page_list", data=payload, lang=lang, static_cache=cache_key)

        cls = models._ENTRY_PAGE_MODELS.get(typing.cast(models.WikiPageType, menu), models.BaseWikiPreview)

        return [cls(**i) for i in data["list"] if i["icon_url"]]

    async def get_wiki_page(
        self,
        id: types.IDOr[models.BaseWikiPreview],
        *,
        lang: typing.Optional[str] = None,
    ) -> models.WikiPage:
        """Get a wiki page."""
        params = dict(entry_page_id=int(id))
        cache_key = cache.cache_key("wiki", endpoint="page", id=id, lang=lang or self.lang)
        data = await self.request_wiki("entry_page", lang=lang, params=params, static_cache=cache_key)

        data["page"].pop("lang", "")  # always an empty string
        return models.WikiPage(**data["page"], lang=lang or self.lang)

    async def get_wiki_pages(
        self,
        ids: typing.Collection[types.IDOr[models.BaseWikiPreview]],
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.WikiPage]:
        """Get multiple wiki pages without modules."""
        payload = dict(entry_page_ids=[int(i) for i in ids])
        data = await self.request_wiki("entry_pages", lang=lang, data=payload)

        return [models.WikiPage(**i) for i in data["entry_pages"]]



"""Web sub client for AuthClient.

Covers HoYoLAB and Miyoushe web auth endpoints.
"""

import json
import typing

import aiohttp

from genshin import constants, errors, types
from genshin.client import routes
from genshin.client.components import base
from genshin.models.auth.cookie import CNWebLoginResult, MobileLoginResult, WebLoginResult
from genshin.models.auth.geetest import SessionMMT, SessionMMTResult
from genshin.utility import auth as auth_utility
from genshin.utility import ds as ds_utility

__all__ = ["WebAuthClient"]


class WebAuthClient(base.BaseClient):
    """Web sub client for AuthClient."""

    @typing.overload
    async def _os_web_login(  # noqa: D102 missing docstring in overload?
        self,
        account: str,
        password: str,
        *,
        encrypted: bool = ...,
        token_type: typing.Optional[int] = ...,
        mmt_result: SessionMMTResult,
    ) -> WebLoginResult: ...

    @typing.overload
    async def _os_web_login(  # noqa: D102 missing docstring in overload?
        self,
        account: str,
        password: str,
        *,
        encrypted: bool = ...,
        token_type: typing.Optional[int] = ...,
        mmt_result: None = ...,
    ) -> typing.Union[SessionMMT, WebLoginResult]: ...

    @base.region_specific(types.Region.OVERSEAS)
    async def _os_web_login(
        self,
        account: str,
        password: str,
        *,
        encrypted: bool = False,
        token_type: typing.Optional[int] = 6,
        mmt_result: typing.Optional[SessionMMTResult] = None,
    ) -> typing.Union[SessionMMT, WebLoginResult]:
        """Login with a password using web endpoint.

        Returns either data from aigis header or cookies.
        """
        headers = {**auth_utility.WEB_LOGIN_HEADERS}
        if mmt_result:
            headers["x-rpc-aigis"] = mmt_result.to_aigis_header()

        payload = {
            "account": account if encrypted else auth_utility.encrypt_credentials(account, 1),
            "password": password if encrypted else auth_utility.encrypt_credentials(password, 1),
            "token_type": token_type,
        }

        async with aiohttp.ClientSession() as session:
            async with session.post(
                routes.WEB_LOGIN_URL.get_url(),
                json=payload,
                headers=headers,
            ) as r:
                data = await r.json()
                cookies = {cookie.key: cookie.value for cookie in r.cookies.values()}

        if data["retcode"] == -3101:
            # Captcha triggered
            aigis = json.loads(r.headers["x-rpc-aigis"])
            return SessionMMT(**aigis)

        if not data["data"]:
            errors.raise_for_retcode(data)

        if data["data"].get("stoken"):
            cookies["stoken"] = data["data"]["stoken"]

        self.set_cookies(cookies)
        return WebLoginResult(**cookies)

    @typing.overload
    async def _cn_web_login(  # noqa: D102 missing docstring in overload?
        self,
        account: str,
        password: str,
        *,
        encrypted: bool = ...,
        mmt_result: SessionMMTResult,
    ) -> CNWebLoginResult: ...

    @typing.overload
    async def _cn_web_login(  # noqa: D102 missing docstring in overload?
        self,
        account: str,
        password: str,
        *,
        encrypted: bool = ...,
        mmt_result: None = ...,
    ) -> typing.Union[SessionMMT, CNWebLoginResult]: ...

    @base.region_specific(types.Region.CHINESE)
    async def _cn_web_login(
        self,
        account: str,
        password: str,
        *,
        encrypted: bool = False,
        mmt_result: typing.Optional[SessionMMTResult] = None,
    ) -> typing.Union[SessionMMT, CNWebLoginResult]:
        """
        Login with account and password using Miyoushe loginByPassword endpoint.

        Returns data from aigis header or cookies.
        """
        headers = {
            **auth_utility.CN_LOGIN_HEADERS,
            "ds": ds_utility.generate_dynamic_secret(constants.DS_SALT["cn_signin"]),
        }
        if mmt_result:
            headers["x-rpc-aigis"] = mmt_result.to_aigis_header()

        payload = {
            "account": account if encrypted else auth_utility.encrypt_credentials(account, 2),
            "password": password if encrypted else auth_utility.encrypt_credentials(password, 2),
        }

        async with aiohttp.ClientSession() as session:
            async with session.post(
                routes.CN_WEB_LOGIN_URL.get_url(),
                json=payload,
                headers=headers,
            ) as r:
                data = await r.json()

        if data["retcode"] == -3102:
            # Captcha triggered
            aigis = json.loads(r.headers["x-rpc-aigis"])
            return SessionMMT(**aigis)

        if not data["data"]:
            errors.raise_for_retcode(data)

        cookies = {cookie.key: cookie.value for cookie in r.cookies.values()}
        self.set_cookies(cookies)

        return CNWebLoginResult(**cookies)

    async def _send_mobile_otp(
        self,
        mobile: str,
        *,
        encrypted: bool = False,
        mmt_result: typing.Optional[SessionMMTResult] = None,
    ) -> typing.Union[None, SessionMMT]:
        """Attempt to send OTP to the provided mobile number.

        May return aigis headers if captcha is triggered, None otherwise.
        """
        headers = {
            **auth_utility.CN_LOGIN_HEADERS,
            "ds": ds_utility.generate_dynamic_secret(constants.DS_SALT["cn_signin"]),
        }
        if mmt_result:
            headers["x-rpc-aigis"] = mmt_result.to_aigis_header()

        payload = {
            "mobile": mobile if encrypted else auth_utility.encrypt_credentials(mobile, 2),
            "area_code": auth_utility.encrypt_credentials("+86", 2),
        }

        async with aiohttp.ClientSession() as session:
            async with session.post(
                routes.MOBILE_OTP_URL.get_url(),
                json=payload,
                headers=headers,
            ) as r:
                data = await r.json()

        if data["retcode"] == -3101:
            # Captcha triggered
            aigis = json.loads(r.headers["x-rpc-aigis"])
            return SessionMMT(**aigis)

        if not data["data"]:
            errors.raise_for_retcode(data)

        return None

    async def _login_with_mobile_otp(self, mobile: str, otp: str, *, encrypted: bool = False) -> MobileLoginResult:
        """Login with OTP and mobile number.

        Returns cookies if OTP matches the one sent, raises an error otherwise.
        """
        headers = {
            **auth_utility.CN_LOGIN_HEADERS,
            "ds": ds_utility.generate_dynamic_secret(constants.DS_SALT["cn_signin"]),
        }

        payload = {
            "mobile": mobile if encrypted else auth_utility.encrypt_credentials(mobile, 2),
            "area_code": auth_utility.encrypt_credentials("+86", 2),
            "captcha": otp,
        }

        async with aiohttp.ClientSession() as session:
            async with session.post(
                routes.MOBILE_LOGIN_URL.get_url(),
                json=payload,
                headers=headers,
            ) as r:
                data = await r.json()

        if not data["data"]:
            errors.raise_for_retcode(data)

        cookies = {cookie.key: cookie.value for cookie in r.cookies.values()}
        self.set_cookies(cookies)

        return MobileLoginResult(**cookies)



"""Genshin lineup models."""

from __future__ import annotations

import datetime
import typing

if typing.TYPE_CHECKING:
    import pydantic.v1 as pydantic
else:
    try:
        import pydantic.v1 as pydantic
    except ImportError:
        import pydantic

from genshin.models.genshin import character
from genshin.models.model import Aliased, APIModel, Unique

__all__ = [
    "Lineup",
    "LineupAbyssScenarios",
    "LineupAbyssScenarios",
    "LineupArtifactStat",
    "LineupArtifactStatFields",
    "LineupCharacter",
    "LineupCharacterPreview",
    "LineupFields",
    "LineupFields",
    "LineupPreview",
    "LineupPrimaryArtifactStat",
    "LineupScenario",
    "LineupWorldScenarios",
    "PartialLineupArtifactSet",
    "PartialLineupCharacter",
    "PartialLineupWeapon",
]


class PartialLineupCharacter(character.BaseCharacter):
    """Partial lineup character."""

    element: str
    rarity: int = Aliased("level")
    weapon_type: str = Aliased("weapon_cat_id")
    icon: str = Aliased("head_icon")
    pc_icon: str = Aliased("standard_icon")

    def __init__(self, _frame: int = 1, **data: typing.Any) -> None:
        # rename "icon" due to pydantic's aliasing
        icons_amount = sum(1 for key, value in data.items() if "icon" in key if value)
        if data.get("pc_icon") and data.get("icon"):
            ...
        elif icons_amount >= 2 and "static" in data.get("icon", ""):
            data["standard_icon"] = data.pop("icon")

        super().__init__(_frame=_frame + 3, **data)  # type: ignore

    @pydantic.validator("element", pre=True)
    def __parse_element(cls, value: typing.Any) -> str:
        if isinstance(value, str) and not value.isdigit():
            return value

        return {
            1: "Pyro",
            2: "Anemo",
            3: "Geo",
            4: "Dendro",
            5: "Electro",
            6: "Hydro",
            7: "Cryo",
        }[int(value)]

    @pydantic.validator("weapon_type", pre=True)
    def __parse_weapon_type(cls, value: typing.Any) -> str:
        if isinstance(value, str) and not value.isdigit():
            return value

        return {
            0: "Unknown",
            1: "Sword",
            10: "Catalyst",
            11: "Claymore",
            12: "Bow",
            13: "Polearm",
        }[int(value)]


class PartialLineupWeapon(APIModel, Unique):
    """Partial lineup weapon."""

    id: int
    name: str
    icon: str
    rarity: int = Aliased("level")
    type: str = Aliased("cat_id")

    @pydantic.validator("type", pre=True)
    def __parse_weapon_type(cls, value: int) -> str:
        if isinstance(value, str) and not value.isdigit():
            return value

        return {
            1: "Sword",
            10: "Catalyst",
            11: "Claymore",
            12: "Bow",
            13: "Polearm",
        }[value]


class PartialLineupArtifactSet(APIModel, Unique):
    """Parital lineup artifact set."""

    id: int
    name: str
    icon: str
    rarity: int = Aliased("level")


class LineupArtifactStatFields(APIModel):
    """Lineup artifact stat fields."""

    flower: typing.Mapping[int, str] = pydantic.Field(artifact_id=1)
    plume: typing.Mapping[int, str] = pydantic.Field(artifact_id=2)
    sands: typing.Mapping[int, str] = pydantic.Field(artifact_id=3)
    goblet: typing.Mapping[int, str] = pydantic.Field(artifact_id=4)
    circlet: typing.Mapping[int, str] = pydantic.Field(artifact_id=5)

    secondary_stats: typing.Mapping[int, str] = Aliased("reliquary_sec_attr")

    @pydantic.root_validator(pre=True)
    def __flatten_stats(cls, values: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
        """Name certain stats."""
        if "reliquary_fst_attr" not in values:
            return values

        artifact_ids = {
            field.field_info.extra["artifact_id"]: name
            for name, field in cls.__fields__.items()
            if field.field_info.extra.get("artifact_id")
        }

        for scenario in values["reliquary_fst_attr"]:
            if scenario["key"] not in artifact_ids:
                continue

            name = artifact_ids[scenario["key"]]
            values[name] = scenario["value"]

        return values

    @pydantic.validator("secondary_stats", "flower", "plume", "sands", "goblet", "circlet", pre=True)
    def __parse_secondary_stats(cls, value: typing.Any) -> typing.Dict[int, str]:
        if not isinstance(value, typing.Sequence):
            return value

        return {stat["id"]: stat["name"] for stat in value}  # type: ignore

    @property
    def all_stats(self) -> typing.Mapping[int, str]:
        """All possible stats for any artifact."""
        return {**self.flower, **self.plume, **self.sands, **self.goblet, **self.circlet, **self.secondary_stats}


class LineupFields(APIModel):
    """Configuration lineup fields."""

    characters: typing.Sequence[PartialLineupCharacter] = Aliased("all_avatar")
    weapons: typing.Sequence[PartialLineupWeapon] = Aliased("all_weapon")
    artifacts: typing.Sequence[PartialLineupArtifactSet] = Aliased("all_set")

    artifact_stats: LineupArtifactStatFields = Aliased("extra_config")


class LineupArtifactStat(APIModel, Unique):
    """Lineup artifact stat."""

    id: int
    name: str


class LineupPrimaryArtifactStat(LineupArtifactStat):
    """Lineup primary artifact stat."""

    artifact_type: int = Aliased("cat_id")


class LineupScenario(APIModel, Unique):
    """Lineup scenario such as domain, boss or spiral abyss."""

    id: int
    name: str
    children: typing.Sequence[LineupScenario]

    @pydantic.root_validator(pre=True)
    def __flatten_scenarios(cls, values: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
        """Name certain scenarios."""
        scenario_ids = {
            field.field_info.extra["scenario_id"]: name
            for name, field in cls.__fields__.items()
            if field.field_info.extra.get("scenario_id")
        }

        for scenario in values["children"]:
            if scenario["id"] not in scenario_ids:
                continue

            name = scenario_ids[scenario["id"]]
            values[name] = scenario

        return values

    @property
    def all_children(self) -> typing.Sequence[LineupScenario]:
        """Get all children of this scenario."""
        children = list(self.children)
        for child in self.children:
            children.extend(child.all_children)

        return children


class LineupWorldScenarios(LineupScenario):
    """Lineup world scenario."""

    trounce_domains: LineupScenario = pydantic.Field(scenario_id=3)
    domain_challenges: LineupScenario = pydantic.Field(scenario_id=9)
    battles: LineupScenario = pydantic.Field(scenario_id=24)


class LineupAbyssScenarios(LineupScenario):
    """Lineup abyss scenario."""

    corridor: LineupScenario = pydantic.Field(scenario_id=42)
    spire: LineupScenario = pydantic.Field(scenario_id=41)


class LineupScenarios(LineupScenario):
    """Lineup scenarios."""

    world: LineupWorldScenarios = pydantic.Field(scenario_id=1)
    abyss: LineupAbyssScenarios = pydantic.Field(scenario_id=2)


class LineupCharacterPreview(PartialLineupCharacter):
    """Partial lineup character preview."""

    role: str = Aliased("avatar_tag")

    icon: str = Aliased("standard_icon")
    pc_icon: str = Aliased("pc_icon")

    @pydantic.validator("role", pre=True)
    def __parse_role(cls, value: typing.Any) -> str:
        if isinstance(value, str):
            return value

        return value["name"]


class LineupCharacter(LineupCharacterPreview):
    """Lineup character."""

    icon: str = Aliased("head_icon")
    pc_icon: str = Aliased("standard_icon")

    weapon: PartialLineupWeapon
    artifacts: typing.Sequence[PartialLineupArtifactSet] = Aliased("set_list")

    artifact_attributes: typing.Sequence[LineupPrimaryArtifactStat] = Aliased("first_attr")
    secondary_attributes: typing.Sequence[LineupArtifactStat] = Aliased("secondary_attr_name")


class LineupPreview(APIModel, Unique):
    """Lineup preview."""

    id: str  # type: ignore
    title: str
    scenario_ids: typing.Sequence[int] = Aliased("tag_ids")

    characters: typing.Sequence[typing.Sequence[LineupCharacter]] = Aliased("avatar_group")

    author_id: int = Aliased("account_uid")
    author_nickname: str = Aliased("nickname")
    author_icon: str = Aliased("avatar_url")
    author_level: int = Aliased("level")

    likes: int = Aliased("like_cnt")
    comments: int = Aliased("comment_cnt")
    created_at: datetime.datetime

    original_lang: str = Aliased("trans_from")

    @pydantic.validator("characters", pre=True)
    def __parse_characters(cls, value: typing.Any) -> typing.Any:
        if isinstance(value[0], typing.Sequence):
            return value

        return [list(group["group"]) for group in value]


class Lineup(LineupPreview):
    """Lineup."""

    description: str
    views: int = Aliased("view_cnt")

    characters: typing.Sequence[typing.Sequence[LineupCharacter]] = Aliased("avatar_group")



"""Genshin serenitea pot replica display models."""

from __future__ import annotations

import datetime
import typing

if typing.TYPE_CHECKING:
    import pydantic.v1 as pydantic
else:
    try:
        import pydantic.v1 as pydantic
    except ImportError:
        import pydantic

from genshin.models.model import Aliased, APIModel, Unique

__all__ = [
    "TeapotReplica",
    "TeapotReplicaAuthor",
    "TeapotReplicaBlueprint",
    "TeapotReplicaStats",
]


class TeapotReplicaAuthor(APIModel):
    """Teapot replica author."""

    id: int = Aliased("user_id")
    nickname: str
    icon: str = Aliased("avatar")


class TeapotReplicaStats(APIModel):
    """Teapot replica stats."""

    diggs: int = Aliased("digg_cnt")
    saves: int = Aliased("store_cnt")
    views: int = Aliased("view_cnt")
    replies: int = Aliased("reply_cnt")
    shares: int = Aliased("share_cnt")
    copies: int = Aliased("copy_cnt")


class TeapotReplicaBlueprint(APIModel):
    """Teapot replica blueprint."""

    share_code: int
    region: str
    module_id: str
    block_id: str
    is_invalid: bool


class TeapotReplica(APIModel, Unique):
    """Genshin serenitea pot replica."""

    post_id: str
    title: str
    content: str
    images: typing.List[str] = Aliased("imgs")
    created_at: datetime.datetime
    stats: TeapotReplicaStats
    lang: str  # type: ignore

    author: TeapotReplicaAuthor

    view_type: int
    sub_type: int
    blueprint: TeapotReplicaBlueprint
    video: typing.Optional[str]

    has_more_content: bool
    token: str

    @pydantic.validator("images", pre=True)
    def __extract_urls(cls, images: typing.Sequence[typing.Any]) -> typing.Sequence[str]:
        return [image if isinstance(image, str) else image["url"] for image in images]

    @pydantic.validator("video", pre=True)
    def __extract_url(cls, video: typing.Any) -> typing.Optional[str]:
        if isinstance(video, str):
            return video

        return video["url"] if video else None



"""Starrail Rogue models."""

from typing import List

from genshin.models.model import APIModel

from ..character import RogueCharacter
from .base import PartialTime


class RogueUserRole(APIModel):
    """Rogue User info."""

    nickname: str
    server: str
    level: int


class RogueBasicInfo(APIModel):
    """generalized rogue basic info."""

    unlocked_buff_num: int
    unlocked_miracle_num: int
    unlocked_skill_points: int


class RogueRecordBasic(APIModel):
    """Basic record info."""

    id: int
    finish_cnt: int
    schedule_begin: PartialTime
    schedule_end: PartialTime


class RogueBuffType(APIModel):
    """Rogue buff type."""

    id: int
    name: str
    cnt: int


class RogueBuffItem(APIModel):
    """Rogue buff item."""

    id: int
    name: str
    is_evoluted: bool
    rank: int


class RogueBuff(APIModel):
    """Rogue buff info."""

    base_type: RogueBuffType
    items: List[RogueBuffItem]


class RogueMiracle(APIModel):
    """Rogue miracle info."""

    id: int
    name: str
    icon: str


class RogueRecordDetail(APIModel):
    """Detailed record info."""

    name: str
    finish_time: PartialTime
    score: int
    final_lineup: List[RogueCharacter]
    base_type_list: List[RogueBuffType]
    cached_avatars: List[RogueCharacter]
    buffs: List[RogueBuff]
    miracles: List[RogueMiracle]
    difficulty: int
    progress: int


class RogueRecord(APIModel):
    """generic record data."""

    basic: RogueRecordBasic
    records: List[RogueRecordDetail]
    has_data: bool


class StarRailRogue(APIModel):
    """generic rogue data."""

    role: RogueUserRole
    basic_info: RogueBasicInfo
    current_record: RogueRecord
    last_record: RogueRecord



import typing

import pytest

from genshin import paginators


class CountingPaginator(paginators.Paginator[int]):
    _index = 0

    async def __anext__(self) -> int:
        if self._index >= 5:
            self._complete()

        self._index += 1
        return self._index


@pytest.fixture(name="counting_paginator")
def counting_paginator_fixture():
    return CountingPaginator()


async def test_paginator_iter(counting_paginator: paginators.Paginator[int]):
    async for value in counting_paginator:
        assert 1 <= value <= 5


async def test_paginator_flatten():
    paginator = CountingPaginator()
    assert await paginator.flatten() == [1, 2, 3, 4, 5]

    paginator = CountingPaginator()
    assert await paginator == [1, 2, 3, 4, 5]


async def test_paginator_next(counting_paginator: paginators.Paginator[int]):
    assert await counting_paginator.next() == 1


async def test_paginator_next_empty():
    paginator = paginators.base.BasicPaginator(())

    with pytest.raises(StopAsyncIteration):
        await paginator.__anext__()

    with pytest.raises(LookupError):
        await paginator.next()


async def test_buffered_paginator():
    class MockBufferedPaginator(paginators.BufferedPaginator[int]):
        async def next_page(self) -> typing.Sequence[int]:
            index = self._counter - 1
            return list(range(index, index + 5))

    paginator = MockBufferedPaginator(limit=12)
    assert not paginator.exhausted

    values = await paginator.flatten()
    assert values == list(range(0, 12))

    assert paginator.exhausted


async def test_merged_paginator():
    # from heapq.merge doc
    sequences = [[1, 3, 5, 7], [0, 2, 4, 8], [5, 10, 15, 20], [], [25]]
    iterators = [paginators.base.aiterate(x) for x in sequences]

    paginator = paginators.MergedPaginator(iterators)
    assert await paginator.flatten() == [0, 1, 2, 3, 4, 5, 5, 7, 8, 10, 15, 20, 25]


async def test_merged_paginator_with_key():
    # from heapq.merge doc
    sequences = [["dog", "horse"], [], ["cat", "fish", "kangaroo"], ["rhinoceros"]]
    iterators = [paginators.base.aiterate(x) for x in sequences]

    paginator = paginators.MergedPaginator(iterators, key=len, limit=5)
    assert await paginator.flatten(lazy=True) == ["dog", "cat", "fish", "horse", "kangaroo"]



"""Genshin chronicle character."""

import typing

if typing.TYPE_CHECKING:
    import pydantic.v1 as pydantic
else:
    try:
        import pydantic.v1 as pydantic
    except ImportError:
        import pydantic

from genshin.models.genshin import character
from genshin.models.model import Aliased, APIModel, Unique

__all__ = [
    "Artifact",
    "ArtifactSet",
    "ArtifactSetEffect",
    "Character",
    "CharacterWeapon",
    "Constellation",
    "Outfit",
    "PartialCharacter",
]


class PartialCharacter(character.BaseCharacter):
    """Character without any equipment."""

    level: int
    friendship: int = Aliased("fetter")
    constellation: int = Aliased("actived_constellation_num")


class CharacterWeapon(APIModel, Unique):
    """Character's equipped weapon."""

    id: int
    icon: str
    name: str
    rarity: int
    description: str = Aliased("desc")
    level: int
    type: str = Aliased("type_name")
    ascension: int = Aliased("promote_level")
    refinement: int = Aliased("affix_level")


class ArtifactSetEffect(APIModel):
    """Effect of an artifact set."""

    pieces: int = Aliased("activation_number")
    effect: str
    enabled: bool = False

    class Config:
        # this is for the "enabled" field, hopefully nobody abuses this
        allow_mutation = True


class ArtifactSet(APIModel, Unique):
    """Artifact set."""

    id: int
    name: str
    effects: typing.Sequence[ArtifactSetEffect] = Aliased("affixes")


class Artifact(APIModel, Unique):
    """Character's equipped artifact."""

    id: int
    icon: str
    name: str
    pos_name: str
    pos: int
    rarity: int
    level: int
    set: ArtifactSet


class Constellation(APIModel, Unique):
    """Character constellation."""

    id: int
    icon: str
    pos: int
    name: str
    effect: str
    activated: bool = Aliased("is_actived")

    @property
    def scaling(self) -> bool:
        """Whether the constellation is simply for talent scaling"""
        return "U" in self.icon


class Outfit(APIModel, Unique):
    """Outfit of a character."""

    id: int
    icon: str
    name: str


class Character(PartialCharacter):
    """Character with equipment."""

    weapon: CharacterWeapon
    artifacts: typing.Sequence[Artifact] = Aliased("reliquaries")
    constellations: typing.Sequence[Constellation]
    outfits: typing.Sequence[Outfit] = Aliased("costumes")

    @pydantic.validator("artifacts")
    def __add_artifact_effect_enabled(cls, artifacts: typing.Sequence[Artifact]) -> typing.Sequence[Artifact]:
        sets: typing.Dict[int, typing.List[Artifact]] = {}
        for arti in artifacts:
            sets.setdefault(arti.set.id, []).append(arti)

        for artifact in artifacts:
            for effect in artifact.set.effects:
                if effect.pieces <= len(sets[artifact.set.id]):
                    effect.enabled = True

        return artifacts



"""Genshin chronicle notes."""

import datetime
import enum
import typing

if typing.TYPE_CHECKING:
    import pydantic.v1 as pydantic
else:
    try:
        import pydantic.v1 as pydantic
    except ImportError:
        import pydantic

from genshin.models.model import Aliased, APIModel

__all__ = [
    "ArchonQuest",
    "ArchonQuestProgress",
    "ArchonQuestStatus",
    "AttendanceReward",
    "AttendanceRewardStatus",
    "DailyTasks",
    "Expedition",
    "Notes",
    "TaskReward",
    "TaskRewardStatus",
]


def _process_timedelta(time: typing.Union[int, datetime.timedelta, datetime.datetime]) -> datetime.datetime:
    if isinstance(time, int):
        time = datetime.datetime.fromtimestamp(time).astimezone()

    if isinstance(time, datetime.timedelta):
        time = datetime.datetime.now().astimezone() + time

    if time < datetime.datetime(2000, 1, 1).astimezone():
        delta = datetime.timedelta(seconds=int(time.timestamp()))
        time = datetime.datetime.now().astimezone() + delta

    time = time.replace(second=0, microsecond=0)

    return time


class Expedition(APIModel):
    """Real-Time note expedition."""

    character_icon: str = Aliased("avatar_side_icon")
    status: typing.Literal["Ongoing", "Finished"]
    remaining_time: datetime.timedelta = Aliased("remained_time")

    @property
    def finished(self) -> bool:
        """Whether the expedition has finished."""
        return self.remaining_time <= datetime.timedelta(0)

    @property
    def completion_time(self) -> datetime.datetime:
        return datetime.datetime.now().astimezone() + self.remaining_time


class TransformerTimedelta(datetime.timedelta):
    """Transformer recovery time."""

    @property
    def timedata(self) -> typing.Tuple[int, int, int, int]:
        seconds: int = super().seconds
        days: int = super().days
        hour, second = divmod(seconds, 3600)
        minute, second = divmod(second, 60)

        return days, hour, minute, second

    @property
    def hours(self) -> int:
        return self.timedata[1]

    @property
    def minutes(self) -> int:
        return self.timedata[2]

    @property
    def seconds(self) -> int:
        return self.timedata[3]


class TaskRewardStatus(str, enum.Enum):
    """Task Reward Statuses."""

    UNFINISHED = "TaskRewardStatusUnfinished"
    FINISHED = "TaskRewardStatusFinished"
    COLLECTED = "TaskRewardStatusTakenAward"


class TaskReward(APIModel):
    """Status of the Commission/Task."""

    status: TaskRewardStatus


class AttendanceRewardStatus(str, enum.Enum):
    """Attendance Reward Statuses."""

    AVAILABLE = "AttendanceRewardStatusWaitTaken"
    COLLECTED = "AttendanceRewardStatusTakenAward"
    FORBIDDEN = "AttendanceRewardStatusForbid"
    UNAVAILABLE = "AttendanceRewardStatusUnfinished"


class AttendanceReward(APIModel):
    """Status of the Encounter Point."""

    status: AttendanceRewardStatus
    progress: int


class DailyTasks(APIModel):
    """Daily tasks section."""

    max_tasks: int = Aliased("total_num")
    completed_tasks: int = Aliased("finished_num")
    claimed_commission_reward: bool = Aliased("is_extra_task_reward_received")

    task_rewards: typing.Sequence[TaskReward]
    attendance_rewards: typing.Sequence[AttendanceReward]
    attendance_visible: bool


class ArchonQuestStatus(str, enum.Enum):
    """Archon quest status."""

    ONGOING = "StatusOngoing"
    NOT_OPEN = "StatusNotOpen"


class ArchonQuest(APIModel):
    """Archon Quest."""

    id: int
    status: ArchonQuestStatus
    chapter_num: str
    chapter_title: str


class ArchonQuestProgress(APIModel):
    """Archon Quest Progress."""

    list: typing.Sequence[ArchonQuest]
    mainlines_finished: bool = Aliased("is_finish_all_mainline")
    archon_quest_unlocked: bool = Aliased("is_open_archon_quest")
    interchapters_finished: bool = Aliased("is_finish_all_interchapter")


class Notes(APIModel):
    """Real-Time notes."""

    current_resin: int
    max_resin: int
    remaining_resin_recovery_time: datetime.timedelta = Aliased("resin_recovery_time")

    current_realm_currency: int = Aliased("current_home_coin")
    max_realm_currency: int = Aliased("max_home_coin")
    remaining_realm_currency_recovery_time: datetime.timedelta = Aliased("home_coin_recovery_time")

    completed_commissions: int = Aliased("finished_task_num")
    max_commissions: int = Aliased("total_task_num")
    claimed_commission_reward: bool = Aliased("is_extra_task_reward_received")

    remaining_resin_discounts: int = Aliased("remain_resin_discount_num")
    max_resin_discounts: int = Aliased("resin_discount_num_limit")

    remaining_transformer_recovery_time: typing.Optional[TransformerTimedelta]

    expeditions: typing.Sequence[Expedition]
    max_expeditions: int = Aliased("max_expedition_num")

    archon_quest_progress: ArchonQuestProgress

    @property
    def resin_recovery_time(self) -> datetime.datetime:
        """The time when resin will be recovered."""
        return datetime.datetime.now().astimezone() + self.remaining_resin_recovery_time

    @property
    def realm_currency_recovery_time(self) -> datetime.datetime:
        """The time when realm currency will be recovered."""
        return datetime.datetime.now().astimezone() + self.remaining_realm_currency_recovery_time

    @property
    def transformer_recovery_time(self) -> typing.Optional[datetime.datetime]:
        """The time the transformer will be recovered."""
        if self.remaining_transformer_recovery_time is None:
            return None

        remaining = datetime.datetime.now().astimezone() + self.remaining_transformer_recovery_time
        return remaining

    @pydantic.root_validator(pre=True)
    def __flatten_transformer(cls, values: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
        if "transformer_recovery_time" in values:
            return values

        if values.get("transformer") and values["transformer"]["obtained"]:
            t = values["transformer"]["recovery_time"]
            delta = TransformerTimedelta(days=t["Day"], hours=t["Hour"], minutes=t["Minute"], seconds=t["Second"])
            values["remaining_transformer_recovery_time"] = delta
        else:
            values["remaining_transformer_recovery_time"] = None

        return values

    daily_task: DailyTasks



"""Game sub client for AuthClient.

Covers OS and CN game auth endpoints.
"""

import json
import typing

import aiohttp

from genshin import constants, errors, types
from genshin.client import routes
from genshin.client.components import base
from genshin.models.auth.cookie import DeviceGrantResult, GameLoginResult
from genshin.models.auth.geetest import RiskyCheckMMT, RiskyCheckMMTResult, RiskyCheckResult
from genshin.models.auth.responses import ShieldLoginResponse
from genshin.utility import auth as auth_utility

__all__ = ["GameAuthClient"]


class GameAuthClient(base.BaseClient):
    """Game sub client for AuthClient."""

    async def _risky_check(
        self, action_type: str, api_name: str, *, username: typing.Optional[str] = None
    ) -> RiskyCheckResult:
        """Check if the given action (endpoint) is risky (whether captcha verification is required)."""
        payload = {"action_type": action_type, "api_name": api_name}
        if username:
            payload["username"] = username

        async with aiohttp.ClientSession() as session:
            async with session.post(
                routes.GAME_RISKY_CHECK_URL.get_url(self.region), json=payload, headers=auth_utility.RISKY_CHECK_HEADERS
            ) as r:
                data = await r.json()

        if not data["data"]:
            errors.raise_for_retcode(data)

        return RiskyCheckResult(**data["data"])

    @typing.overload
    async def _shield_login(  # noqa: D102 missing docstring in overload?
        self,
        account: str,
        password: str,
        *,
        encrypted: bool = ...,
        mmt_result: RiskyCheckMMTResult,
    ) -> ShieldLoginResponse: ...

    @typing.overload
    async def _shield_login(  # noqa: D102 missing docstring in overload?
        self,
        account: str,
        password: str,
        *,
        encrypted: bool = ...,
        mmt_result: None = ...,
    ) -> typing.Union[ShieldLoginResponse, RiskyCheckMMT]: ...

    async def _shield_login(
        self,
        account: str,
        password: str,
        *,
        encrypted: bool = False,
        mmt_result: typing.Optional[RiskyCheckMMTResult] = None,
    ) -> typing.Union[ShieldLoginResponse, RiskyCheckMMT]:
        """Log in with the given account and password.

        Returns MMT if geetest verification is required.
        """
        if self.default_game is None:
            raise ValueError("No default game set.")

        headers = auth_utility.SHIELD_LOGIN_HEADERS.copy()
        if mmt_result:
            headers["x-rpc-risky"] = mmt_result.to_rpc_risky()
        else:
            # Check if geetest is required
            check_result = await self._risky_check("login", "/shield/api/login", username=account)
            if check_result.mmt:
                return check_result.to_mmt()
            else:
                headers["x-rpc-risky"] = auth_utility.generate_risky_header(check_result.id)

        payload = {
            "account": account,
            "password": password if encrypted else auth_utility.encrypt_credentials(password, 2),
            "is_crypto": True,
        }
        async with aiohttp.ClientSession() as session:
            async with session.post(
                routes.SHIELD_LOGIN_URL.get_url(self.region, self.default_game), json=payload, headers=headers
            ) as r:
                data = await r.json()

        if not data["data"]:
            errors.raise_for_retcode(data)

        return ShieldLoginResponse(**data["data"])

    @typing.overload
    async def _send_game_verification_email(  # noqa: D102 missing docstring in overload?
        self,
        action_ticket: str,
        *,
        mmt_result: RiskyCheckMMTResult,
    ) -> None: ...

    @typing.overload
    async def _send_game_verification_email(  # noqa: D102 missing docstring in overload?
        self,
        action_ticket: str,
        *,
        mmt_result: None = ...,
    ) -> typing.Union[None, RiskyCheckMMT]: ...

    async def _send_game_verification_email(
        self, action_ticket: str, *, mmt_result: typing.Optional[RiskyCheckMMTResult] = None
    ) -> typing.Union[None, RiskyCheckMMT]:
        """Send email verification code.

        Returns `None` if success, `RiskyCheckMMT` if geetest verification is required.
        """
        headers = auth_utility.GRANT_TICKET_HEADERS.copy()
        if mmt_result:
            headers["x-rpc-risky"] = mmt_result.to_rpc_risky()
        else:
            # Check if geetest is required
            check_result = await self._risky_check("device_grant", "/device/api/preGrantByTicket")
            if check_result.mmt:
                return check_result.to_mmt()
            else:
                headers["x-rpc-risky"] = auth_utility.generate_risky_header(check_result.id)

        payload = {
            "way": "Way_Email",
            "action_ticket": action_ticket,
            "device": {
                "device_model": "iPhone15,4",
                "device_id": auth_utility.DEVICE_ID,
                "client": 1,
                "device_name": "iPhone",
            },
        }
        async with aiohttp.ClientSession() as session:
            async with session.post(
                routes.PRE_GRANT_TICKET_URL.get_url(self.region), json=payload, headers=headers
            ) as r:
                data = await r.json()

        if data["retcode"] != 0:
            errors.raise_for_retcode(data)

        return None

    async def _verify_game_email(self, code: str, action_ticket: str) -> DeviceGrantResult:
        """Verify the email code."""
        payload = {"code": code, "ticket": action_ticket}
        async with aiohttp.ClientSession() as session:
            async with session.post(
                routes.DEVICE_GRANT_URL.get_url(self.region), json=payload, headers=auth_utility.GRANT_TICKET_HEADERS
            ) as r:
                data = await r.json()

        return DeviceGrantResult(**data["data"])

    @base.region_specific(types.Region.OVERSEAS)
    async def _os_game_login(self, uid: str, game_token: str) -> GameLoginResult:
        """Log in to the game."""
        if self.default_game is None:
            raise ValueError("No default game set.")

        payload = {
            "channel_id": 1,
            "device": auth_utility.DEVICE_ID,
            "app_id": constants.APP_IDS[self.default_game][self.region],
        }
        payload["data"] = json.dumps({"uid": uid, "token": game_token, "guest": False})
        payload["sign"] = auth_utility.generate_sign(payload, constants.APP_KEYS[self.default_game][self.region])

        async with aiohttp.ClientSession() as session:
            async with session.post(
                routes.GAME_LOGIN_URL.get_url(self.region, self.default_game),
                json=payload,
                headers=auth_utility.GAME_LOGIN_HEADERS,
            ) as r:
                data = await r.json()

        if not data["data"]:
            errors.raise_for_retcode(data)

        return GameLoginResult(**data["data"])



"""Starrail chronicle challenge."""

from typing import TYPE_CHECKING, Any, Dict, List, Optional

if TYPE_CHECKING:
    import pydantic.v1 as pydantic
else:
    try:
        import pydantic.v1 as pydantic
    except ImportError:
        import pydantic

from genshin.models.model import Aliased, APIModel
from genshin.models.starrail.character import FloorCharacter

from .base import PartialTime

__all__ = [
    "FictionBuff",
    "FictionFloor",
    "FictionFloorNode",
    "FloorNode",
    "StarRailChallenge",
    "StarRailFloor",
    "StarRailPureFiction",
]


class FloorNode(APIModel):
    """Node for a floor."""

    challenge_time: PartialTime
    avatars: List[FloorCharacter]


class StarRailFloor(APIModel):
    """Floor in a challenge."""

    name: str
    round_num: int
    star_num: int
    node_1: FloorNode
    node_2: FloorNode
    is_chaos: bool


class StarRailChallenge(APIModel):
    """Challenge in a season."""

    season: int = Aliased("schedule_id")
    begin_time: PartialTime
    end_time: PartialTime

    total_stars: int = Aliased("star_num")
    max_floor: str
    total_battles: int = Aliased("battle_num")
    has_data: bool

    floors: List[StarRailFloor] = Aliased("all_floor_detail")


class FictionBuff(APIModel):
    """Buff for a Pure Fiction floor."""

    id: int
    name: str = Aliased("name_mi18n")
    description: str = Aliased("desc_mi18n")
    icon: str


class FictionFloorNode(FloorNode):
    """Node for a Pure Fiction floor."""

    buff: Optional[FictionBuff]
    score: int


class FictionFloor(APIModel):
    """Floor in a Pure Fiction challenge."""

    id: int = Aliased("maze_id")
    name: str
    round_num: int
    star_num: int
    node_1: FictionFloorNode
    node_2: FictionFloorNode
    is_fast: bool

    @property
    def score(self) -> int:
        """Total score of the floor."""
        return self.node_1.score + self.node_2.score


class StarRailPureFiction(APIModel):
    """Pure Fiction challenge in a season."""

    name: str
    season_id: int
    begin_time: PartialTime
    end_time: PartialTime

    total_stars: int = Aliased("star_num")
    max_floor: str
    total_battles: int = Aliased("battle_num")
    has_data: bool

    floors: List[FictionFloor] = Aliased("all_floor_detail")
    max_floor_id: int

    @pydantic.root_validator(pre=True)
    def __unnest_groups(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if "groups" in values and isinstance(values["groups"], List):
            groups: List[Dict[str, Any]] = values["groups"]
            if len(groups) > 0:
                values["name"] = groups[0]["name_mi18n"]
                values["season_id"] = groups[0]["schedule_id"]
                values["begin_time"] = groups[0]["begin_time"]
                values["end_time"] = groups[0]["end_time"]

        return values



"""Utility functions related to genshin."""

import typing

from genshin import types

__all__ = [
    "create_short_lang_code",
    "get_prod_game_biz",
    "recognize_game",
    "recognize_genshin_server",
    "recognize_honkai_server",
    "recognize_region",
    "recognize_server",
    "recognize_starrail_server",
]

UID_RANGE: typing.Mapping[types.Game, typing.Mapping[types.Region, typing.Sequence[str]]] = {
    types.Game.GENSHIN: {
        types.Region.OVERSEAS: ("6", "7", "8", "18", "9"),
        types.Region.CHINESE: ("1", "2", "3", "5"),
    },
    types.Game.STARRAIL: {
        types.Region.OVERSEAS: ("6", "7", "8", "9"),
        types.Region.CHINESE: ("1", "2", "5"),
    },
    types.Game.HONKAI: {
        types.Region.OVERSEAS: ("1", "2"),
        types.Region.CHINESE: ("3", "4"),
    },
}
"""Mapping of games and regions to their respective UID ranges."""

GENSHIN_SERVER_RANGE: typing.Mapping[str, typing.Sequence[str]] = {
    "cn_gf01": ("1", "2", "3"),
    "cn_qd01": ("5",),
    "os_usa": ("6",),
    "os_euro": ("7",),
    "os_asia": ("8", "18"),
    "os_cht": ("9",),
}
"""Mapping of Genshin servers to their respective UID ranges."""

STARRAIL_SERVER_RANGE: typing.Mapping[str, typing.Sequence[str]] = {
    "prod_gf_cn": ("1", "2"),
    "prod_qd_cn": ("5",),
    "prod_official_usa": ("6",),
    "prod_official_eur": ("7",),
    "prod_official_asia": ("8",),
    "prod_official_cht": ("9",),
}
"""Mapping of Star Rail servers to their respective UID ranges."""


def create_short_lang_code(lang: str) -> str:
    """Create an alternative short lang code."""
    return lang if "zh" in lang else lang.split("-")[0]


def recognize_genshin_server(uid: int) -> str:
    """Recognize which server a Genshin UID is from."""
    for server_name, digits in GENSHIN_SERVER_RANGE.items():
        if str(uid)[:-8] in digits:
            return server_name

    raise ValueError(f"UID {uid} isn't associated with any server")


def get_prod_game_biz(region: types.Region, game: types.Game) -> str:
    """Get the game_biz value corresponding to a game and region."""
    game_biz = ""
    if game == types.Game.HONKAI:
        game_biz = "bh3_"
    elif game == types.Game.GENSHIN:
        game_biz = "hk4e_"
    elif game == types.Game.STARRAIL:
        game_biz = "hkrpg_"

    if region == types.Region.OVERSEAS:
        game_biz += "global"
    elif region == types.Region.CHINESE:
        game_biz += "cn"

    return game_biz


def recognize_honkai_server(uid: int) -> str:
    """Recognizes which server a Honkai UID is from."""
    if 10000000 < uid < 100000000:
        return "overseas01"
    elif 100000000 < uid < 200000000:
        return "usa01"
    elif 200000000 < uid < 300000000:
        return "eur01"

    # From what I can tell, CN UIDs are all over the place,
    # seemingly even overlapping with overseas UIDs...
    # Probably gonna need some input from actual CN players here, but I know none.
    # It could be that e.g. global range is 2e8 ~ 2.5e8
    raise ValueError(f"UID {uid} isn't associated with any server")


def recognize_starrail_server(uid: int) -> str:
    """Recognize which server a Star Rail UID is from."""
    for server, digits in STARRAIL_SERVER_RANGE.items():
        if str(uid)[:-8] in digits:
            return server

    raise ValueError(f"UID {uid} isn't associated with any server")


def recognize_server(uid: int, game: types.Game) -> str:
    """Recognizes which server a UID is from."""
    if game == types.Game.HONKAI:
        return recognize_honkai_server(uid)
    elif game == types.Game.GENSHIN:
        return recognize_genshin_server(uid)
    elif game == types.Game.STARRAIL:
        return recognize_starrail_server(uid)
    else:
        raise ValueError(f"{game} is not a valid game")


def recognize_game(uid: int, region: types.Region) -> typing.Optional[types.Game]:
    """Recognize the game of a uid."""
    if len(str(uid)) == 8:
        return types.Game.HONKAI

    for game, digits in UID_RANGE.items():
        if str(uid)[:-8] in digits[region]:
            return game

    return None


def recognize_region(uid: int, game: types.Game) -> typing.Optional[types.Region]:
    """Recognize the region of a uid."""
    for region, digits in UID_RANGE[game].items():
        if str(uid)[:-8] in digits:
            return region

    return None



"""Starrail Chronicle Base Model."""

import datetime

from genshin.models.model import APIModel


class PartialTime(APIModel):
    """Partial time model."""

    year: int
    month: int
    day: int
    hour: int
    minute: int

    @property
    def datetime(self) -> datetime.datetime:
        return datetime.datetime(self.year, self.month, self.day, self.hour, self.minute)



"""CLI tools."""

import asyncio
import datetime
import functools
import http.cookies
import os
import typing

import click

import genshin

if os.name == "nt":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())  # type: ignore


T = typing.TypeVar("T", bound=typing.Any)

cli: click.Group = click.Group("cli")


def asynchronous(func: typing.Callable[..., typing.Awaitable[typing.Any]]) -> typing.Callable[..., typing.Any]:
    """Make an asynchronous function runnable by click."""

    @functools.wraps(func)
    def wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        return asyncio.run(func(*args, **kwargs))  # type: ignore # unsure what is wrong here

    return wrapper


def client_command(func: typing.Callable[..., typing.Awaitable[typing.Any]]) -> typing.Callable[..., typing.Any]:
    """Make a click command that uses a Client."""

    @click.option("--cookies", help="Cookie header to use in authentication.", default=None)
    @click.option("--lang", help="Default language.", default="en-us")
    @click.option("--debug", is_flag=True, hidden=True)
    @functools.wraps(func)
    @asynchronous
    async def command(
        cookies: typing.Optional[str] = None,
        lang: str = "en-us",
        debug: bool = False,
        **kwargs: typing.Any,
    ) -> typing.Any:
        client = genshin.Client(cookies, lang=lang, debug=debug)
        if cookies is None:
            client.set_browser_cookies()

        return await func(client, **kwargs)

    return command


@cli.command()
@client_command
async def accounts(client: genshin.Client) -> None:
    """Get all of your genshin accounts."""
    data = await client.get_game_accounts()

    for account in data:
        click.echo(
            f"[{account.game_biz}] {click.style(str(account.uid), bold=True)} - "
            f"{account.nickname} lvl {account.level} ({account.server_name})"
        )


genshin_group: click.Group = click.Group("genshin", help="Genshin-related commands.")
honkai_group: click.Group = click.Group("honkai", help="Honkai-related commands.")
starrail_group: click.Group = click.Group("starrail", help="StarRail-related commands.")
cli.add_command(genshin_group)
cli.add_command(honkai_group)
cli.add_command(starrail_group)


@honkai_group.command("stats")
@click.argument("uid", type=int)
@client_command
async def honkai_stats(client: genshin.Client, uid: int) -> None:
    """Show simple honkai statistics."""
    cuid = click.style(str(uid), fg="blue")
    click.echo(f"User stats of {cuid}\n")

    data = await client.get_honkai_user(uid)

    click.secho("Stats:", fg="yellow")
    for k, v in data.stats.as_dict(lang=client.lang).items():
        if isinstance(v, dict):
            click.echo(f"{k}:")
            for nested_k, nested_v in typing.cast("typing.Dict[str, object]", v).items():
                click.echo(f"  {nested_k}: {click.style(str(nested_v), bold=True)}")
        else:
            click.echo(f"{k}: {click.style(str(v), bold=True)}")


@genshin_group.command("stats")
@click.argument("uid", type=int)
@client_command
async def genshin_stats(client: genshin.Client, uid: int) -> None:
    """Show simple genshin statistics."""
    cuid = click.style(str(uid), fg="blue")
    click.echo(f"User stats of {cuid}\n")

    data = await client.get_partial_genshin_user(uid)

    click.secho("Stats:", fg="yellow")
    for k, v in data.stats.as_dict(lang=client.lang).items():
        value = click.style(str(v), bold=True)
        click.echo(f"{k}: {value}")

    click.echo()
    click.secho("Explorations:", fg="yellow")
    for area in data.explorations:
        perc = click.style(str(area.explored) + "%", bold=True)
        offerings = ", ".join(f"{o.name} {click.style(str(o.level), bold=True)}" for o in area.offerings)
        click.echo(f"{area.name} - explored {perc} | {offerings}")

    if data.teapot is not None:
        click.echo()
        click.secho("Teapot:", fg="yellow")
        level = click.style(str(data.teapot.level), bold=True)
        comfort = click.style(str(data.teapot.comfort), bold=True)
        click.echo(f"level {level} | comfort {comfort} ({data.teapot.comfort_name})")
        click.echo(f"Unlocked realms: {', '.join(r.name for r in data.teapot.realms)}")


@genshin_group.command("characters")
@click.argument("uid", type=int)
@client_command
async def genshin_characters(client: genshin.Client, uid: int) -> None:
    """Show genshin characters."""
    cuid = click.style(str(uid), fg="blue")
    click.echo(f"Characters of {cuid}")

    characters = await client.get_genshin_characters(uid)
    characters = sorted(characters, key=lambda c: (c.level, c.rarity), reverse=True)

    for char in characters:
        color = {
            "Anemo": "bright_green",
            "Pyro": "red",
            "Hydro": "bright_blue",
            "Electro": "magenta",
            "Cryo": "bright_cyan",
            "Geo": "yellow",
            "Dendro": "green",
        }.get(char.element)

        click.echo()
        name = click.style(char.name, bold=True)
        element = click.style(char.element, fg=color)
        click.echo(f"{name} ({'★' * char.rarity} {element})")
        click.echo(f"lvl {char.level} C{char.constellation}, friendship lvl {char.friendship}")
        click.echo(
            f"Weapon: {char.weapon.name} ({'★' * char.weapon.rarity} {char.weapon.type}) - "
            f"lvl {char.weapon.level} R{char.weapon.refinement}"
        )

        if char.artifacts:
            click.echo("Artifacts:")
            for arti in char.artifacts:
                click.echo(f" - {arti.pos_name}: {arti.set.name} ({'★' * arti.rarity})")

        if char.outfits:
            click.echo(f"Outfits: {', '.join(o.name for o in char.outfits)}")


@genshin_group.command("notes")
@click.argument("uid", type=int, default=None, required=False)
@client_command
async def genshin_notes(client: genshin.Client, uid: typing.Optional[int]) -> None:
    """Show real-Time notes."""
    click.echo("Real-Time notes.")

    data = await client.get_notes(uid)

    click.echo(f"{click.style('Resin:', bold=True)} {data.current_resin}/{data.max_resin}")
    click.echo(f"{click.style('Realm currency:', bold=True)} {data.current_realm_currency}/{data.max_realm_currency}")
    click.echo(
        f"{click.style('Commissions:', bold=True)} " f"{data.completed_commissions}/{data.max_commissions}",
        nl=False,
    )
    if data.completed_commissions == data.max_commissions and not data.claimed_commission_reward:
        click.echo(f" | [{click.style('X', fg='red')}] Haven't claimed rewards")
    else:
        click.echo()
    click.echo(
        f"{click.style('Used resin cost-halving opportunities:', bold=True)} "
        f"{data.max_resin_discounts - data.remaining_resin_discounts}/{data.max_resin_discounts}"
    )
    if data.transformer_recovery_time and data.transformer_recovery_time > datetime.datetime.now(datetime.timezone.utc):
        click.echo(f"{click.style('Transformer recovery:', bold=True)} {data.transformer_recovery_time}")

    click.echo(f"\n{click.style('Expeditions:', bold=True)} {len(data.expeditions)}/{data.max_expeditions}")
    for expedition in data.expeditions:
        if expedition.remaining_time > datetime.timedelta(0):
            remaining = f"{expedition.remaining_time} remaining"
            click.echo(f" - {expedition.status} | {remaining}")
        else:
            click.echo(f" - {expedition.status}")


@starrail_group.command("notes")
@click.argument("uid", type=int, default=None, required=False)
@client_command
async def starrail_notes(client: genshin.Client, uid: typing.Optional[int]) -> None:
    """Show real-Time starrail notes."""
    click.echo("Real-Time notes.")

    data = await client.get_starrail_notes(uid)

    click.echo(f"{click.style('TB power:', bold=True)} {data.current_stamina}/{data.max_stamina}", nl=False)
    click.echo(f" (Full in {data.stamina_recover_time})" if data.stamina_recover_time > datetime.timedelta(0) else "")
    click.echo(f"{click.style('Reserved TB power:', bold=True)} {data.current_reserve_stamina}/2400")
    click.echo(f"{click.style('Daily training:', bold=True)} {data.current_train_score}/{data.max_train_score}")
    click.echo(f"{click.style('Simulated Universe:', bold=True)} {data.current_rogue_score}/{data.max_rogue_score}")
    click.echo(
        f"{click.style('Echo of War:', bold=True)} {data.remaining_weekly_discounts}/{data.max_weekly_discounts}"
    )

    click.echo(f"\n{click.style('Assignments:', bold=True)} {data.accepted_epedition_num}/{data.total_expedition_num}")
    for expedition in data.expeditions:
        if expedition.remaining_time > datetime.timedelta(0):
            remaining = f"{expedition.remaining_time} remaining"
            click.echo(f" - {expedition.name} | {remaining}")
        else:
            click.echo(f" - {expedition.name} | Finished")


@cli.command()
@click.option("--scenario", help="Scenario ID or name to use (eg '12-3').", type=str, default=None)
@client_command
async def lineups(client: genshin.Client, scenario: typing.Optional[str]) -> None:
    """Show popular genshin lineups."""
    scenarios = await client.get_lineup_scenarios()

    if scenario is None:
        scenario_id = None
    elif scenario.isdigit():
        scenario_id = int(scenario)
    else:
        for s in scenarios.all_children:
            if s.name.lower() == scenario.lower():
                scenario_id = s.id
                break
        else:
            raise Exception(f"Scenario {scenario!r} not found.")

    lineups = await client.get_lineups(scenario_id, limit=10, page_size=10)

    for lineup in lineups:
        scenario_names = [s.name for s in scenarios.all_children if s.id in lineup.scenario_ids]

        click.echo(f"{click.style(lineup.title, fg='cyan')} | {click.style(str(lineup.likes), fg='green')} likes")
        click.echo(f"{click.style('Author:', bold=True)} {lineup.author_nickname} (AR {lineup.author_level})")
        click.echo(f"{click.style('Scenarios:', bold=True)} {', '.join(scenario_names)}")
        click.echo(f"{click.style('Characters:', bold=True)}")
        for group, characters in enumerate(lineup.characters, 1):
            if len(lineup.characters) > 1:
                click.echo(f"- Group {group}:")
            for character in characters:
                click.echo(f" - {character.name} ({character.role})")

        click.echo("")


@cli.command()
@click.option("--limit", help="The maximum amount of wishes to show.", type=int, default=None)
@client_command
async def wishes(client: genshin.Client, limit: typing.Optional[int] = None) -> None:
    """Show a nicely formatted wish history."""
    client.set_authkey()

    banner_names = await client.get_banner_names()
    longest = max(len(v) for v in banner_names.values())

    async for wish in client.wish_history(limit=limit):
        banner = click.style(wish.banner_name.ljust(longest), bold=True)
        click.echo(f"{banner} | {wish.time.astimezone()} - {wish.name} ({'★' * wish.rarity} {wish.type})")


@cli.command()
@client_command
async def pity(client: genshin.Client) -> None:
    """Calculate the amount of pulls until pity."""
    client.set_authkey()

    banners = await client.get_banner_names()
    for banner, name in banners.items():
        # skip the novice banner
        if banner == 100:
            continue

        click.echo()
        click.secho(name, fg="yellow")

        accum = 0
        async for wish in client.wish_history(banner):
            accum += 1
            if wish.rarity == 5:
                a = click.style(str(90 - accum), bold=True)
                n = click.style(wish.name, fg="green")
                click.secho(f"Pulled {n} {accum} pulls ago. {a} pulls left until pity.")
                break
        else:
            a = click.style(str(90 - accum), bold=True)
            click.secho(f"Never pulled a 5*. At most {a} pulls left until pity")


@cli.command(hidden=True)
@client_command
async def banner_ids(client: genshin.Client) -> None:
    """Get the banner ids from logs."""
    ids = genshin.utility.get_genshin_banner_ids()

    click.echo("\n".join(ids) + "\n")

    if len(ids) < 3:
        click.echo("Please open all detail pages!")
        return

    for banner in await client.get_banner_details(ids):
        click.echo(f"{banner.banner_id} - {banner.banner_type_name} ({banner.banner_type})")


@cli.command(hidden=True)
def authkey() -> None:
    """Get an authkey from logfiles."""
    click.echo(genshin.utility.get_authkey())


@cli.command()
@click.option("-a", "--account", default=None, prompt=True)
@click.option("-p", "--password", default=None, prompt=True, hide_input=True)
@click.option("--port", help="Webserver port.", type=int, default=5000)
@asynchronous
async def login(account: str, password: str, port: int) -> None:
    """Login with a password."""
    client = genshin.Client()
    result = await client.os_login_with_password(account, password, port=port)
    cookies = await genshin.complete_cookies(result.dict())

    base: http.cookies.BaseCookie[str] = http.cookies.BaseCookie(cookies)
    click.echo(f"Your cookies are: {click.style(base.output(header='', sep=';'), bold=True)}")


if __name__ == "__main__":
    cli()



"""Ratelimit handlers."""

import asyncio
import functools
import typing

from genshin import errors

CallableT = typing.TypeVar("CallableT", bound=typing.Callable[..., typing.Awaitable[typing.Any]])


def handle_ratelimits(
    tries: int = 5,
    exception: typing.Type[errors.GenshinException] = errors.VisitsTooFrequently,
    delay: float = 0.3,
) -> typing.Callable[[CallableT], CallableT]:
    """Handle ratelimits for requests."""
    # TODO: Support exponential backoff

    def wrapper(func: typing.Callable[..., typing.Awaitable[typing.Any]]) -> typing.Any:
        @functools.wraps(func)
        async def inner(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
            for _ in range(tries):
                try:
                    x = await func(*args, **kwargs)
                except exception:
                    await asyncio.sleep(delay)
                else:
                    return x
            else:
                raise exception({}, f"Got ratelimited {tries} times in a row")

        return inner

    return wrapper
